
class signature(tuple):
    def __new__(cls, *args, **kwargs):
        return super().__new__( cls, (cls.__name__, args, kwargs) )
    def __init__(self, *args, **kwargs):
        super().__init__()



import os
import winshell
import string
from .config import *
from ..smarts import smartmatch, SmartMatch
import re
from typing import List, Type
import sys

PUNTUACTION = string.punctuation.replace('-','')
PUNTUACTION_RE_ESCAPED = re.escape(PUNTUACTION)
TRANSLATION_TABLE = str.maketrans(PUNTUACTION, ":"*len(PUNTUACTION))

class Pattern:
    id = None
    re1 = None
    base_path = None
    zone = None
    subs_sep = ''
    def __init__(self, path=None):
        super().__init__()
        self.path = path
        self.matched = None
    @classmethod
    def includes(cls, path):
        return path.startswith(cls.base_path)   
    def match(self):
        self.matched = smartmatch()
        return self.matched(self.re1,self.path)  
    def test(self, spec):
        return re.match(self.re1,spec)     
    def canonical(self, default=None, compact=False):
        if self.matched:
            code = default
            zone = self.matched[('zone',str,None)]
            if zone is None:
                return code
            code = zone
            root = self.matched[('root',str,None)]
            if root is not None:
                code = f'{code}-{root}'
                subs = self.matched[('subs',str,None)]
                if subs is not None:
                    code = f'{code}{self.subs_sep}{subs.translate(TRANSLATION_TABLE)}'
            if not compact:
                name = self.matched[('name',str,None)]
                if name is not None:
                    code = f'{code}-{name}'
            return code
        else:
            rel_path = self.path[len(self.base_path)+1:]
            parts = rel_path.split(os.path.sep)
            for part in parts[:1]:
                if re.match(self.re1, part):
                    code = self.canonicalize(part,None,True)
                else:
                    code = f'{self.zone}-{part}'
            if len(parts)>1:
                for part in parts[1:-1]:
                    part_code = self.canonicalize(part,part,True)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
                for part in parts[-1:]:
                    part_code = self.canonicalize(part,part,compact)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
            return code
        
    @classmethod
    def canonicalize(cls, spec, default=None, compact=False):
        if os.path.sep in spec:
            real_path = os.path.realpath(spec)            
            if not cls.includes(real_path):
                return default
            return cls(real_path).canonical(default, compact)
        else:
            p = cls(spec)
            if not p.match():
                return default
            return p.canonical(default, compact)

class Pattern11(Pattern):
    id = '11'
    zone = '11'
    subs_sep = ''
    re1 = r'(?P<zone>11)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\1'

class Pattern21(Pattern):
    id = '21'
    zone = '21'
    subs_sep = ''
    re1 = r'(?P<zone>21)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\2'

class Pattern31(Pattern):
    zone = '31'
    subs_sep = ''
    def canonical(self, default=None, compact=False):
        if self.matched:
            code = default
            zone = self.matched[('zone',str,None)]
            if zone is None:
                return code
            code = zone
            root = self.matched[('root',str,None)]
            if root is not None:
                code = f'{code}-{root}'
                subs = self.matched[('subs',str,None)]
                if subs is not None:
                    code = f'{code}{self.subs_sep}{subs.translate(TRANSLATION_TABLE)}'
            if not compact:
                name = self.matched[('name',str,None)]
                if name is not None:
                    code = f'{code}-{name}'
            return code
        else:
            rel_path = self.path[len(self.base_path)+1:]
            parts = rel_path.split(os.path.sep)
            for part in parts[:1]:
                if re.match(self.re1, part):
                    code = self.canonicalize(part,None,True)
                else:
                    code = f'31-{part}'
            if len(parts)>1:
                for part in parts[1:-1]:
                    part_code = self.canonicalize(part,part,True)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
                for part in parts[-1:]:
                    part_code = self.canonicalize(part,part,compact)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
            return code

class Pattern31d(Pattern31):
    id = '31d'
    re1 = r'(?P<zone>31)-(?P<root>D[A-Z0-9]{4})(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\3\\1\\1'

class Pattern31m(Pattern31):
    id = '31m'
    re1 = r'(?P<zone>31)-(?P<root>M[0-9]{2})(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\3\\1\\2'
    subs_sep = ':'

class Pattern31o(Pattern31):
    id = '31o'
    re1 = r'(?P<zone>31)-(?P<root>O[A-Z0-9]{4})(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\3\\1\\3'

class Pattern31c(Pattern31):
    id = '31c'
    re1 = r'(?P<zone>31)-(?P<root>C[0-9]{2})(?:(?P<subs>[^-]+))?(?:-(?P<name>.+))?'
    base_path = 'C:\\1\\W\\3\\1\\4'
    subs_sep = ':'


class Pattern41(Pattern):
    id = '41'
    zone = '41'
    re_name = r'(?P<name>.+)'
    re_subs = r'(?P<subs>[^-]+)'
    re_time = r'(?P<time>[0-9]{4})'
    re_day = r'(?P<day>[0-9]{2})'
    re_month = r'(?P<month>[0-9]{2})'
    re_year = r'(?P<year>[0-9]{2})'
    re_zone = r'(?P<zone>41)'
    re1 = rf'{re_zone}-{re_year}(?:{re_month}(?:{re_day}(?:\.{re_time}(?:{re_subs})?)?)?)?(?:-{re_name})?'
    base_path = 'C:\\1\\W\\4'
    #format_code = lambda self, zone, root, subs: f'{zone}-{root}:{subs}'
    def canonical(self, default=None, compact=False):
        if self.matched:
            code = default
            zone = self.matched[('zone',str,None)]
            if zone is None:
                return code
            code = zone
            year = self.matched[('year',str,None)]
            if year is not None:
                code = f'{code}-{year}'
                month = self.matched[('month',str,None)]
                if month is not None:
                    code = f'{code}:{month}'
                    day = self.matched[('day',str,None)]
                    if day is not None:
                        code = f'{code}:{day}'
                        time = self.matched[('time',str,None)]
                        if time is not None:
                            code = f'{code}:{time}'
                            subs = self.matched[('subs',str,'')]
                            if subs is not None:
                                code = f'{code}:{subs.translate(TRANSLATION_TABLE)}'
            if not compact:
                name = self.matched[('name',str,None)]
                if name is not None:
                    code = f'{code}-{name}'
            return code
        else:
            rel_path = self.path[len(self.base_path)+1:]
            parts = rel_path.split(os.path.sep)
            for part in parts[:1]:
                if re.match(self.re1, part):
                    code = self.canonicalize(part,None,True)
                else:
                    code = f'{self.zone}-{part}'
            if len(parts)>1:
                for part in parts[1:-1]:
                    part_code = self.canonicalize(part,part,True)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
                for part in parts[-1:]:
                    part_code = self.canonicalize(part,part,compact)
                    if part_code.startswith(code):
                        code = part_code
                    else:
                        code += f':{part}'
            return code

class Pattern51(Pattern):
    id = '51'
    zone = '51'
    subs_sep = ''
    re1 = r'(?P<zone>51)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.*))?'
    base_path = 'C:\\1\\W\\5'


class Pattern61(Pattern):
    id = '61'
    zone = '61'
    subs_sep = ''
    re1 = r'(?P<zone>61)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.*))?'
    base_path = 'C:\\1\\W\\6'

class Pattern71(Pattern):
    id = '71'
    zone = '71'
    subs_sep = ''
    re1 = r'(?P<zone>71)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.*))?'
    base_path = 'C:\\1\\W\\7'

class PatternD1(Pattern):
    id = 'D1'
    zone = 'D1'
    subs_sep = ''
    re1 = r'(?P<zone>D1)-(?P<root>[A-Z0-9]+)(?:(?P<subs>[^-]+))?(?:-(?P<name>.*))?'
    base_path = 'C:\\1\\Y\\D1'

patterns : List[Type[Pattern]]= [
    Pattern11,
    Pattern21,
    Pattern31d,
    Pattern31m,
    Pattern31o,
    Pattern31c,
    Pattern41,
    Pattern51,
    Pattern61,
    Pattern71,
    PatternD1
]

'''
def canonical_name(path, default=None, compact=False):
    if os.path.sep in path:
        real_path = os.path.realpath(path)
        for pat in patterns:
            if pat.includes(real_path):
                break
        else:
            return default
        return pat(path).canonical(default, compact)
    else:
        for pat in patterns:
            p = pat(path)
            if p.match():
                break
        else:
            return default
        return p.canonical(default, compact)
'''

def pathname_complete(spec:str, default=None, options=[]):
    for pat in patterns:
        p = pat(spec)
        if p.match():
            break
    else:
        return default    
    spec_c = p.canonical(default)
    if spec_c is default:
        return spec_c
    spec_p = spec_c.split('-')
    if len(spec_p)<2:
        if os.path.exists(p.base_path):
            return p.base_path
        return
    result_list = []
    row_index = 0
    def walker(base_path, spec_r):   
        nonlocal row_index
        row_index += 1    
        print(f'>>> {row_index:5}: {base_path}', file=sys.stderr)
        matches = 0
        ld = os.listdir(base_path)
        spec_k = spec_c
        ld_c = []
        for item in ld:
            path = os.path.join(base_path,item)
            item_c = p.canonicalize(path)
            if item_c is None:
                continue 
            ld_c.append((item,item_c,path))
        ld_p = []
        ref_len = len(spec_r)
        while len(spec_k)>ref_len:
            for item, item_c, path in ld_c:
                if item_c.startswith(spec_k):
                    matches += 1
                    if spec_k == spec_c:
                        result_list.append(path)
                    if os.path.isdir(path):
                        ld_p.append((path,spec_k))
            if matches:
                break
            #spec_k = ':'.join(spec_k.split(':')[:-1]) if spec_k.count('-')==1 else '-'.join(spec_k.split('-')[0:2])
            spec_k = spec_k[:-1]
        for path, spec_k in ld_p:
            matches += walker(path,spec_k)
        return matches
    if not os.path.exists(p.base_path):
        return None
    walker(p.base_path,'')
    if len(result_list)==0:
        return None
    if 'resolve-links' in options:
        final_list = []
        for result_path in result_list:
            result_stack = [result_path]
            link_target = result_path
            while link_target.endswith('.lnk'):
                try:
                    with winshell.shortcut(result_path) as link:
                        link_target = link.path
                    #print('link-target-2', link_target)
                    if not os.path.exists(link_target):
                        break
                    result_stack.append(link_target)
                except:
                    break
            if 'stacked' in options:
                final_list.append(result_stack)
            else:
                final_list.append(result_stack[-1])
    else:
        if 'stacked' in options:
            final_list = [[c] for c in result_list]
        else:
            final_list = result_list
    return final_list if 'as-list' in options else final_list[0]



def locate(doc:str, options=[]):
    return pathname_complete(doc,options)






'''

def ___canonical_name(path, alt=None, compact=False):
    prefix = ''
    if path.find(os.path.sep)>=0:
        path_real = os.path.realpath(path)
        for part, area, drive in ( (W_FOLDER_PATH,'W','X:\\W'), 
                                  (S_FOLDER_PATH,'S','X:\\S'), 
                                  (X_FOLDER_PATH,'X','X:\\X'), 
                                  (Y_FOLDER_PATH,'Y','Y:\\') ):        
            if path_real.startswith(part):
                mark = 3
                break
            if path_real.startswith(drive):
                mark = 2
                break
        else:
            return alt
        path_parts = path_real.split(os.path.sep)
        if mark>=len(path_parts):
            return alt
        match area:
            case 'W':
                zone = path_parts[mark] + '1'
            case 'S':
                zone = path_parts[mark] + '0'
            case 'Y':
                zone = '-'.join(path_parts[mark] .split('-')[:2])
            case _:
                return alt
        offset = mark + 1
        if offset>=len(path_parts):
            return zone
        prefix=f'{zone}-'
        if zone=='31':
            if path_parts[offset] != '1':
                return alt
            offset += 1
            if offset>=len(path_parts):
                return alt
            map = {'1':'D','2':'M','3':'O','4':'F','5':'C','6':'S'}
            if path_parts[offset] not in map:
                return alt
            offset += 1
            if offset>=len(path_parts):
                return alt
            part = path_parts[offset]
            p = __canonical_name(part,'',compact=True)
        else:
            part = path_parts[offset]
            p = __canonical_name(part,'',compact=True)
        if p.startswith(zone):
            prefix = p
        else:
            for c in part:
                if c in string.punctuation:
                    return alt
            prefix += part
        offset += 1 
        if offset>=len(path_parts):
            return prefix 
        offset += 1
        for i, part in enumerate(path_parts[offset:-1],0):
            p = __canonical_name(part,'',compact=True)
            if p.startswith(prefix):
                prefix = p
            else:
                for c in part:
                    if c in string.punctuation:
                        return alt
                prefix = f'{prefix}:{part}' 
        part = path_parts[-1]
        p = __canonical_name(part,'')
        if p.startswith(prefix):
            return p
        else:
            for c in part:
                if c in string.punctuation:
                    return alt
            return f'{prefix}:{part}'
    else:
        parts = path.split('-')
        if 2>len(parts):
            return alt
        parts[1] = ''.join([c if c not in PUNTUACTION else ':' for c in parts[1] ])
        if parts[0][0] == '4':
            if len(parts[1])>2:
                parts[1] = parts[1][:2] + ':'+ parts[1][2:]    
            if len(parts[1])>5:
                parts[1] = parts[1][:5] + ':'+ parts[1][5:]    
        if compact:
            return '-'.join(parts[:2])
        else:
            return '-'.join(parts)


'''



# layer or tier?

class Layer:
    def __init__(self):
        pass
    def item_build(self, builder):
        pass
    def item_stage(self, builder):
        pass
    def item_bind(self, builder):
        pass
    def item_enter(self, builder):
        pass
    def item_leave(self, builder):
        pass
    def __enter__(self): 
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class CommonLayer(Layer):
    def item_build(self, builder):
        this = None 
        if type(builder.type) is type:
            init = builder.kwargs.get('init', None)
            if init:
                method, args = init 
                this = builder.type(**args)
            else:
                this = builder.type()
        else:
            if type(builder.item) is type:
                this = builder.item()     
        return this



from ctypes.wintypes import *
import ctypes
from sortedcontainers import SortedDict
from . import gui11x0_wintraits as wintraits
#from . import gui13x0_winapi as winapi

class STARTUPINFOEX(ctypes.Structure):
    _fields_ = [  ('cb', DWORD) ,
                    ('lpReserved', LPSTR),  
                    ('lpDesktop', LPSTR),
                    ('lpTitle', LPSTR),
                    ('dwX', DWORD),
                    ('dwy', DWORD),
                    ('dwXSize', DWORD),
                    ('dwYSize', DWORD),
                    ('dwXCountChars', DWORD),
                    ('dwYCountChars', DWORD),
                    ('dwFillAttribute', DWORD),
                    ('dwFlags', DWORD),
                    ('wShowWindow', WORD),
                    ('cbReserved2', WORD),
                    ('lpReserved2', LPBYTE),
                    ('hStdInput', HANDLE),
                    ('hStdOutput', HANDLE),
                    ('hStdError', HANDLE),
                    ('lpAttributeList', ctypes.POINTER(ctypes.c_void_p)) ]

class PROCESS_INFORMATION (ctypes.Structure):
    _fields_ = [('hProcess', HANDLE),
                    ('hThread', HANDLE),
                    ('dwProcessId',DWORD),
                    ('dwThreadId',DWORD)]

GUID = ctypes.c_ubyte * 16

class PROPERTYKEY(ctypes.Structure):
    _fields_ = [("fmtid", GUID),
                ("pid", DWORD)]

VARIANT_BOOL = USHORT
class VARIANT_VALUE(ctypes.Union):
    _fields_ = [("boolVal", VARIANT_BOOL), 
                ("pwszVal", LPWSTR)]

class PROPVARIANT(ctypes.Structure):
    VT_EMPTY	        = 0
    VT_NULL     	    = 1
    VT_I2	            = 2
    VT_I4	            = 3
    VT_R4	            = 4
    VT_R8	            = 5
    VT_CY	            = 6
    VT_DATE	            = 7
    VT_BSTR	            = 8
    VT_DISPATCH	        = 9
    VT_ERROR	        = 10
    VT_BOOL	            = 11
    VT_VARIANT	        = 12
    VT_UNKNOWN	        = 13
    VT_DECIMAL	        = 14
    VT_I1	            = 16
    VT_UI1	            = 17
    VT_UI2	            = 18
    VT_UI4	            = 19
    VT_I8	            = 20
    VT_UI8	            = 21
    VT_INT	            = 22
    VT_UINT	            = 23
    VT_VOID	            = 24
    VT_HRESULT	        = 25
    VT_PTR	            = 26
    VT_SAFEARRAY	    = 27
    VT_CARRAY	        = 28
    VT_USERDEFINED	    = 29
    VT_LPSTR	        = 30
    VT_LPWSTR	        = 31
    VT_RECORD	        = 36
    VT_INT_PTR	        = 37
    VT_UINT_PTR	        = 38
    VT_FILETIME	        = 64
    VT_BLOB	            = 65
    VT_STREAM	        = 66
    VT_STORAGE	        = 67
    VT_STREAMED_OBJECT	= 68
    VT_STORED_OBJECT	= 69
    VT_BLOB_OBJECT	    = 70
    VT_CF	            = 71
    VT_CLSID	        = 72
    VT_VERSIONED_STREAM	= 73
    VT_BSTR_BLOB	    = 0xfff
    VT_VECTOR	        = 0x1000
    VT_ARRAY	        = 0x2000
    VT_BYREF	        = 0x4000
    VT_RESERVED	        = 0x8000
    VT_ILLEGAL	        = 0xffff
    VT_ILLEGALMASKED	= 0xfff
    VT_TYPEMASK	        = 0xfff

    _fields_ = [("vt", USHORT),
                ("wReserved1", USHORT),
                ("wReserved2", USHORT),
                ("wReserved3", USHORT),
                ("vVal", VARIANT_VALUE)]

class IPropertyStoreVtbl(ctypes.Structure):
    _fields_ = [
        ('QueryInterface', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.POINTER(GUID), ctypes.POINTER(ctypes.c_void_p))),
        ('AddRef', ctypes.CFUNCTYPE(ctypes.c_ulong, ctypes.c_void_p)),
        ('Release', ctypes.CFUNCTYPE(ctypes.c_ulong, ctypes.c_void_p)),
        ('GetCount', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.POINTER(ctypes.c_ulong))),
        ('GetAt', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.c_ulong, ctypes.POINTER(PROPERTYKEY))),
        ('GetValue', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.POINTER(PROPERTYKEY), ctypes.POINTER(PROPVARIANT))),
        ('SetValue', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.POINTER(PROPERTYKEY), ctypes.POINTER(PROPVARIANT))),
        ('Commit', ctypes.CFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p))
    ]

class IPropertyStore(ctypes.Structure):
    # {886D8EEB-8CF2-4446-8D02-CDBA1DBDCF99}
    IID_IPropertyStore = (GUID)(*bytearray.fromhex("eb8e6d88f28c46448d02cdba1dbdcf99"))
    _fields_ = [('lpVtbl', ctypes.POINTER(IPropertyStoreVtbl))]

class PropertyStock:
    # {9F4C2855-9F79-4B39-A8D0-E1D42DE1D5F3}
    PKEY_AppUserModel = (GUID)(*bytearray.fromhex("55284c9f799f394ba8d0e1d42de1d5f3"))
    catalog = SortedDict(
        {
            wintraits.System.AppUserModel.ID: (PKEY_AppUserModel, 5, str, PROPVARIANT.VT_LPWSTR),
            wintraits.System.AppUserModel.RelaunchCommand: (PKEY_AppUserModel, 2, str, PROPVARIANT.VT_LPWSTR),
            wintraits.System.AppUserModel.RelaunchDisplayNameResource: (PKEY_AppUserModel, 4, str, PROPVARIANT.VT_LPWSTR),
            wintraits.System.AppUserModel.RelaunchIconResource: (PKEY_AppUserModel, 3, str, PROPVARIANT.VT_LPWSTR),
            wintraits.System.AppUserModel.PreventPinning: (PKEY_AppUserModel, 9, bool, PROPVARIANT.VT_BOOL)
        }
    )

LRESULT = ctypes.c_long
WNDPROC = ctypes.WINFUNCTYPE(ctypes.c_long, HWND, UINT, WPARAM, LPARAM)


from . import types
from . import wintypes
from . import winapi
from . import wintools
from . import wintraits
import time
import psutil
from sortedcontainers import SortedDict, SortedSet
from typing import Tuple
from PIL import Image, ImageChops
from PIL.Image import Image as PilImage


def is_in_color_range(px: Tuple[int, int, int], minimal_color:int) -> bool:
    return px[0] >= minimal_color and px[1] >= minimal_color and px[2] >= minimal_color


def is_line_color_range(im: PilImage, y: int, minimal_color: int) -> bool:
    width, _ = im.size

    for x in range(0, width-1):
        px = im.getpixel((x, y))
        if not is_in_color_range(px, minimal_color):
            return False
    return True


def is_column_color_range(im: PilImage, x: int, minimal_color: int) -> bool:
    _, height = im.size

    for y in range(0, height-1):
        px = im.getpixel((x, y))
        if not is_in_color_range(px, minimal_color):
            return False
    return True

def get_crop_box_by_px_color(
    im: PilImage, 
    px: Tuple[int, int], 
    scale: float, 
    offset: int) -> Tuple[int, int, int, int]:

    bg = Image.new(im.mode, im.size, px)
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, scale, offset)
    return diff.getbbox()

def crop_by_background(
    im: PilImage, 
    minimal_light_background_color_value: int) -> Tuple[int, int, int, int]:

    width, height = im.size
    original_box = (0, 0, width, height)

    # crop by top left pixel color
    px = im.getpixel((0,height-1))
    if is_in_color_range(px, minimal_light_background_color_value):
        bbox1 = get_crop_box_by_px_color(im, px, 2.0, -100)
    else: 
        bbox1 = original_box

    # crop by bottom right pixel color
    px = im.getpixel((width-1,height-1))
    if is_in_color_range(px, minimal_light_background_color_value):
        bbox2 = get_crop_box_by_px_color(im, px, 2.0, -100)
    else:
        bbox2 = original_box

    crop = (
        max(bbox1[0], bbox2[0]),
        max(bbox1[1], bbox2[1]),
        min(bbox1[2], bbox2[2]),
        min(bbox1[3], bbox2[3])
    )

    return crop


def list_windows(options=None) -> list[types.W]:
    return types.W.list_windows(options)

def rake_windows(options=None, **kwargs):
    return types.W.rake_windows(options,**kwargs)



from pyrite.gui import tools, wintools, winapi
import rich
from rich.table import Table
import re
import sys
import argparse

def main():    
    winapi.CoInitialize() # needed for wintools.hwnd_system_app_user_model_id_get(w.hwnd)
    parser = argparse.ArgumentParser("window-rake.py")
    parser.add_argument("--title", "-t", help="Item name fragment", type=str)
    parser.add_argument("--pname", "-n", help="Item name fragment", type=str)
    parser.add_argument("--pexe", "-x", help="Item name fragment", type=str)
    parser.add_argument("--cmdline", "-c", help="Item name fragment", type=str)
    args = parser.parse_args()
    arg_list = ['title','pname','pexe','cmdline']
    print(f'--title=[{args.title.__repr__()}]', file=sys.stderr)
    wl = tools.rake_windows(**{k:args.__dict__[k] for k in arg_list if args.__dict__[k] is not None})
    if wl:
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Z", style="dim")
        #table.add_column("alias")
        table.add_column("scode", overflow="fold")
        table.add_column("title", overflow="fold")
        table.add_column("model", overflow="fold")
        table.add_column("pname", overflow="fold")
        table.add_column("cmdline", overflow="fold")
        table.add_column("id")
        table.add_column("pid")
        table.add_column("class", overflow="fold")
        for w in wl:
            scode = w.spread_code_guess(options="complete")
            table.add_row(
                w.ztag, #w.w_shortcut, 
                scode, 
                w.title.encode('unicode_escape').decode(), 
                wintools.hwnd_system_app_user_model_id_get(w.hwnd), 
                w.pname, 
                ' '.join(w.pcmdline),
                str(w.hwnd), 
                str(w.pid),
                w.classname
            )
        rich.print(table)


if __name__ == '__main__':
    main()
    


import re

class AnsiEscapeParser:
    pattern = re.compile(r"""
    (
        \x1b(?:
            (?P<CSI>\[[\x30-\x3F]*[\x20-\x2F]*[\x40-\x7E])   #CSI
            |
            (?P<OSC>\][^\x07\x9C]*(?:[\x07\x9C]|\x1b\x5C))   #OSC
            |
            (?P<Fs>[\x60-\x7E])                              #Fs
            |
            (?P<Fp>[\x30-\x3F])                              #Fp
            |                               
            (?P<nF>[\x20-\x2F]*[\x30-\x7E])                  #nF
            |
            (?P<partial>.*$)                                 #unterminated ESC
        )
        |
        (?P<text>[^\x1b]+)                                   #normal text
    )
    """, re.VERBOSE)
    def __init__(self):
        self.sequence_order = 0
        self.pending_text = ''
        self.last_item = None
        pass
    def parse(self, text):
        all_text = self.pending_text + text
        self.pending_text = ''
        pending_text = ''
        for i, match in enumerate(self.pattern.finditer(all_text), self.sequence_order+1):
            token = match.group()
            kind = tuple([k for k,v in match.groupdict().items() if v is not None])            
            if 'text' in kind:
                token = pending_text + token
                pending_text = ''
                if token != '' and self.last_item is not None:
                    if token=='\r\n' and self.last_item[2]=='\x1b[K':
                        self.last_item = (i,kind,token)
                        continue
                    if 'CSI' in self.last_item[1] and self.last_item[2].endswith(';1H'):
                        token = '\r\n' + token
                if token.endswith('\r'):
                    pending_text = '\r'
                    token = token[:-1]            
            self.last_item = (i,kind,token)
            yield self.last_item
        self.pending_text = pending_text
        if match.groupdict()['partial'] is not None:
            self.pending_text += token
            self.sequence_order = i - 1
        else:
            self.sequence_order = i  
    def reset(self):
        self.sequence_order = 0
        self.pending_text = ''
        self.last_item = None
        pass



from tkinter import Misc
from typing import Any, Literal
from . import gui01x0_tktraits as tktraits
from . import gui02x0_tktypes as tktypes
from . import gui03x0_tktools as tktools
from .gui04x1_tkliterals import *

class VerticalScrolledFrame(tktypes.Frame):
    class MyCanvas(tktypes.Canvas):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args,**kwargs)
        def yview(self,*args):
            super().yview(*args)
        def yview_scroll(self,*args,**kwargs):
            super().yview_scroll(*args,**kwargs)
            
        def yview_moveto(self,*args,**kwargs):
            super().yview_moveto(*args,**kwargs)

    class MyScrollbar(tktypes.Scrollbar):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args,**kwargs)
        def set(self,first,last):
            super().set(first,last)
            f, l = self.get()
            if f <= 0.0 or l >= 1.0:
                self.event_generate('<<scroll_limit_reached>>', when="tail")


    def __init__(self, master, *args, **kw):
        tktypes.Frame.__init__(self, master, *args, **kw)

        self.cw = 0
        self.sw = 0
        self.sh = 0
        self.inner_frame = tktypes.Frame(self)
        self.inner_frame.pack(side=tkside.TOP, fill=tkfill.X)
        self.inner_frame.pack_propagate(False)
        self.pack_propagate(False)

        # Create a canvas object and a vertical scrollbar for scrolling it.
        vscrollbar = self.MyScrollbar(self.inner_frame, orient=tkorient.VERTICAL)
        vscrollbar.pack(side=tkside.RIGHT, fill=tkfill.Y, expand=tkexpand.NO, anchor=tkanchor.N)
        self.canvas = tktypes.Canvas(self.inner_frame, bd=0, highlightthickness=0, 
                                width = 0, height = 0,
                                yscrollcommand=vscrollbar.set)
        self.canvas.pack(side=tkside.LEFT, fill=tkfill.BOTH, expand=tkexpand.YES, anchor=tkanchor.N)
        #self.canvas.pack_propagate(False)
        vscrollbar.config(command = self.canvas.yview)
        self.vscrollbar = vscrollbar
        # Reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # Create a frame inside the canvas which will be scrolled with it.
        self.interior = tktypes.Frame(self.canvas,height=0)
        self.interior.bind('<Configure>', self._configure_interior)
        self.canvas.bind('<Configure>', self._configure_canvas)
        self.bind('<Configure>', self._configure_inner_frame)
        #self.vscrollbar.bind('<<scroll_limit_reached>>', self._scroll_limit_reached)
        self.interior_id = self.canvas.create_window(0, 0, window=self.interior, anchor=tkanchor.NW)
        self.after_id = None
        self.last_scroll = (-1,-1)

        self.rebind()

    def scroll(self,target):
        if target == 'top':
            self.canvas.yview_moveto(0)

    def _scroll_limit_reached(self, event):
        return
        lfirst, llast = self.last_scroll
        first, last = self.vscrollbar.get()
        l, t, r, b = self.canvas.bbox('all')
        ih =  self.inner_frame.winfo_height()
        h = b - t
        print('_scroll_limit_reached: ', (first, last), (lfirst, llast), (l,t,r,b), ih)
        self.last_scroll = (first,last)
        if first<=0:         
            if self.after_id is not None:
                self.after_cancel(self.after_id)  
            self.after_id = self.after(300,lambda self=self:self.canvas.yview_scroll(24/h, 'units'))
        elif last>=1:
            if self.after_id is not None:
                self.after_cancel(self.after_id)  
            self.after_id = self.after(300,lambda self=self:self.canvas.yview_scroll(-24/h,'units'))

    def __refresh(self, height=None):
        if height is None:
            height = self.winfo_height()
        bbl, bbt, bbr, bbb = self.canvas.bbox("all")
        #self.canvas.config(scrollregion=(bbl,bbt-17,bbr,bbb+16))
        bbh = bbb - bbt
        h = height
        if h > bbh:
            h = bbh
            self.canvas.config(scrollregion=(bbl,bbt,bbr,bbb))
        else:
            #self.canvas.config(scrollregion=(bbl,bbt-17,bbr,bbb+16))
            self.canvas.config(scrollregion=(bbl,bbt-25,bbr,bbb+24))
        self.inner_frame.config(height=h)
        #print(f'VerticalScrolledFrame.__refresh: bbl={bbl}, bbt={bbt}, bbr={bbr}, bbb={bbb}, bbh={bbh}, h={h}')

    def _configure_inner_frame(self, event):
        if event.widget.winfo_id() == self.winfo_id():
            self.__refresh(event.height)
            #bbl, bbt, bbr, bbb = self.canvas.bbox("all")
            #bbh = bbb - bbt
            #h = event.height
            #if h > bbh:
            #    h = bbh
            #self.inner_frame.config(height=h)

    def _configure_interior(self, event):
        #print(f'_configure_interior {event.widget.winfo_name()} ~ {self.winfo_name()}')
        #self.canvas.config(scrollregion=self.canvas.bbox("all"))
        self.__refresh()

    def _configure_canvas(self, event):
        #print(f'_configure_canvas {event.widget.winfo_name()} ~ {self.winfo_name()}')
        self.canvas.itemconfigure(self.interior_id, width=self.canvas.winfo_width())

    def on_mouse_wheel(self, event):
        self.canvas.yview_scroll(int(-event.delta/120),"units")
        return "break" 
    
    def rebind(self):
        def recursive_bind(widget,event,handler):
            widget.bind(event,handler)
            for ww in widget.winfo_children():
                recursive_bind(ww,event,handler)
        recursive_bind(self.interior,"<MouseWheel>", self.on_mouse_wheel)





#from pywinauto.win32defines import *
from win32con import *

import ctypes

#define DPI_AWARENESS_CONTEXT_UNAWARE              ((DPI_AWARENESS_CONTEXT)-1)
#define DPI_AWARENESS_CONTEXT_SYSTEM_AWARE         ((DPI_AWARENESS_CONTEXT)-2)
#define DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE    ((DPI_AWARENESS_CONTEXT)-3)
#define DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2 ((DPI_AWARENESS_CONTEXT)-4)
#define DPI_AWARENESS_CONTEXT_UNAWARE_GDISCALED    ((DPI_AWARENESS_CONTEXT)-5)

DPI_AWARENESS_CONTEXT_UNAWARE = -1
DPI_AWARENESS_CONTEXT_SYSTEM_AWARE = -2
DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE = -3
DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2 = -4
DPI_AWARENESS_CONTEXT_UNAWARE_GDISCALED = -5

EXTENDED_STARTUPINFO_PRESENT = 0x00080000

class System:
    class AppUserModel:
        ID = 'System.AppUserModel.ID'
        RelaunchCommand = 'System.AppUserModel.RelaunchCommand'
        RelaunchIconResource = 'System.AppUserModel.RelaunchIconResource'
        RelaunchDisplayNameResource = 'System.AppUserModel.RelaunchDisplayNameResource'
        PreventPinning = 'System.AppUserModel.PreventPinning'










from .gui11x0_wintraits import *
from . import gui12x0_wintypes as wintypes
import ctypes
__PROC_THREAD_ATTRIBUTE_PARENT_PROCESS = 0x00020000
PROC_THREAD_ATTRIBUTE_PARENT_PROCESS = ctypes.cast(__PROC_THREAD_ATTRIBUTE_PARENT_PROCESS,ctypes.POINTER(wintypes.DWORD))


from . import gui16x1_winapi as winapi
from . import gui15x1_wintypes as wintypes
from . import gui14x1_wintraits as wintraits
from typing import Tuple
from sortedcontainers import SortedSet
import re
import os
import psutil
from PIL import Image, ImageTk, ImageGrab, ImageChops
import PIL.ImageDraw
import PIL
import xml.etree.ElementTree as ET
import ctypes
import time

def hmon_list_monitors(options=None):
    monitor_list = []
    for index, (hmon, hdc, (l, t, r, b)) in enumerate(winapi.EnumDisplayMonitors(),1):
        if options is not None:
            monitor_list.append((hmon, hdc, (l, t, r, b)))
        else:
            monitor_list.append(hmon)
    return monitor_list

def hmon_from_monitor_number(monitor_number):
    mon_list = hmon_list_monitors()
    if monitor_number<1 or monitor_number>len(mon_list):
        return None
    return mon_list[monitor_number-1]

def hmon_from_hwnd(hwnd):
    return winapi.MonitorFromWindow(hwnd, wintraits.MONITOR_DEFAULTTONEAREST)

# ltrb stands for left-top-right-bottom
def hmon_from_rect_ltbr(ltrb_tuple):
    return winapi.MonitorFromRect(ltrb_tuple, wintraits.MONITOR_DEFAULTTONEAREST)

def hwnd_restore(hwnd):
    try:
        cpid = winapi.GetCurrentProcessId()
        fghwnd = winapi.GetForegroundWindow()
        fgtid, fgpid = winapi.GetWindowThreadProcessId(fghwnd)
        tid, pid = winapi.GetWindowThreadProcessId(hwnd)
        if pid==cpid and fgpid==cpid:
            #print('hwnd_window_show: trying internally...')
            if winapi.IsIconic(hwnd):
                winapi.ShowWindow(hwnd, wintraits.SW_RESTORE)
            winapi.SetForegroundWindow(hwnd)
            time.sleep(0.1)
            if winapi.SetForegroundWindow(hwnd) == hwnd:
                return
    except Exception as e:
        print('exception in hwnd_window_show:', e)
    
    try:
        #print('hwnd_window_show: trying with external process...')
        winapi.WinExec(fr'C:\1\X\tools\windows-spread-switcher\build\windows-spread-switcher.exe {hwnd} show')
        return
    except Exception as e:
        print('exception in hwnd_window_show:', e)


def hwnd_minimize(hwnd):
    #print(f'hwnd_minimize {hwnd}')
    try:
        cpid = winapi.GetCurrentProcessId()
        tid, pid = winapi.GetWindowThreadProcessId(hwnd)
        if pid==cpid:
            #print('hwnd_minimize: trying internally...')
            if winapi.IsIconic(hwnd):
                winapi.ShowWindow(hwnd,wintraits.SW_RESTORE)
            else:
                winapi.ShowWindow(hwnd,wintraits.SW_SHOWMINIMIZED)
            return
    except Exception as e:
        print('exception in hwnd_window_show:', e)

    try:
        #print('hwnd_minimize: trying with external process...')
        winapi.WinExec(fr'C:\1\X\tools\windows-spread-switcher\build\windows-spread-switcher.exe {hwnd} minimize')
        return
    except Exception as e:
        print('exception in hwnd_minimize:', e)

def hwnd_close(hwnd):
    #print(f'hwnd_close {hwnd}')
    try:
        winapi.PostMessage(hwnd,wintraits.WM_CLOSE,0,0);
    except Exception as e:
        print('exception in hwnd_window_show:', e)

def hwnd_move(hwnd, x, y, w, h, repaint):
    #print('hwnd_move_window')
    #return win32gui.SetWindowPos(hwnd,win32con.NULL,x,y,w,h,win32con.SWP_ASYNCWINDOWPOS) #win32con.SWP_NOACTIVATE|win32con.SWP_NOZORDER|(win32con.SWP_NOREDRAW if not repaint else 0))
    winapi.MoveWindow(hwnd,x,y,w,h,repaint)

def hwnd_zorder(hwnd):
    z_order = -1
    hwnd2 = hwnd
    while hwnd2 != wintraits.NULL:        
        hwnd2 = winapi.GetWindow(hwnd2, wintraits.GW_HWNDPREV)
        z_order += 1
    return z_order

#@warnings.deprecated('deprecata perchè fuorviante. usare hwnd_toplevel_from_tk_window o hwnd_toplevel_from_tk_widget')
def hwnd_from_tk_window(tk_wwindow):
    hwnd = tk_wwindow.winfo_id()
    hwnd_parent = hwnd
    while hwnd_parent != wintraits.NULL:
        hwnd = hwnd_parent
        hwnd_parent = winapi.GetParent(hwnd)
    return hwnd

def hwnd_toplevel_from_tk_window(tk_window):
    hwnd = tk_window.winfo_id()
    hwnd_parent = hwnd
    while hwnd_parent != wintraits.NULL:
        hwnd = hwnd_parent
        hwnd_parent = winapi.GetParent(hwnd)
    return hwnd

def hwnd_toplevel_from_tk_widget(tk_widget):
    hwnd = tk_widget.winfo_id()
    hwnd_parent = hwnd
    while hwnd_parent != wintraits.NULL:
        hwnd = hwnd_parent
        hwnd_parent = winapi.GetParent(hwnd)
    return hwnd

def hwnd_list_child_windows(hwnd):
    @staticmethod
    def win32_window_load_children_cb(hwnd_child, w_set):
        w_set.append(hwnd_child)
    w_set = []
    winapi.EnumChildWindows(hwnd, win32_window_load_children_cb, w_set)
    return w_set

def hwnd_grab_icon(hwnd, tpid:Tuple[int, int]=None):
        try:
            wt = winapi.GetWindowText(hwnd)
            if wt == 'Settings':
                pass
            if tpid is None:
                tid, pid = winapi.GetWindowThreadProcessId(hwnd)
            else:
                tid, pid = tpid
            
            #if wt == 'Settings':
            #    win32_get_gui_thread_info non serve ad un mazza
            #    Window.win32_get_gui_thread_info(tid)
            p = psutil.Process(pid)
            if p.name()=='ApplicationFrameHost.exe':
                try:
                    # questo sotto tenta di recuperare il processo hostato dentro ApplicationFrameHost
                    # però funziona solo se la finestra è visibile
                    cwl = hwnd_list_child_windows(hwnd)             
                    for cw in cwl:
                        ctid, cpid = winapi.GetWindowThreadProcessId(cw)
                        if cpid != pid:
                            p = psutil.Process(cpid)
                            break
                except Exception as e:
                    print('Exception ', e)
            try:
                cp_dir = os.path.split(p.exe())[0]
                appx_manifest_fn = os.path.join(cp_dir, 'AppxManifest.xml')
                for i in range(1,3):
                    if os.path.exists(appx_manifest_fn):
                        break
                    cp_dir = os.path.split(cp_dir)[0]
                    appx_manifest_fn = os.path.join(cp_dir, 'AppxManifest.xml')
                if os.path.exists(appx_manifest_fn):
                    tree = ET.parse(appx_manifest_fn)
                    root = tree.getroot()
                    ns = {"win10":"http://schemas.microsoft.com/appx/manifest/foundation/windows10"}
                    # pathToLogo = manifest.Root.Element(XName.Get("Properties", ns)).Element(XName.Get("Logo", ns)).Value;
                    properties = root.find('win10:Properties', namespaces=ns)
                    logo = properties.find("win10:Logo", namespaces=ns)
                    #print(logo.text)
                    logo_fn = os.path.join(cp_dir,logo.text)
                    #print(logo_fn, os.path.exists(logo_fn))
                    fdn = os.path.split(logo_fn)
                    fne = os.path.splitext(fdn[1])
                    spec = f'{fdn[0]}\\{fne[0]}*{fne[1]}'
                    ls = os.listdir(fdn[0])
                    lx = SortedSet()
                    for ff in ls:
                        if re.match(fr'{fne[0]}.*\{fne[1]}',ff):
                            lx.add(ff)
                    if logo_fn not in lx:
                        index = lx.bisect_right(f'{fdn[1]}.scale-')
                        if index>=len(lx):
                            return None
                        logo_fn = lx[index]

                    #print(f'{p.name()} [{wt}] [{logo_fn}]')
                    logo_final_fn = os.path.join(fdn[0],logo_fn)
                    img = PIL.Image.open(logo_final_fn)
                    width, height = img.size
                    if height>24:
                        img = img.resize((width*24//height, 24))
                    image = ImageTk.PhotoImage(img)
                    return image
            except Exception as e:
                print('Exception ', e)
                pass
                

            # hicon = win32gui.GetClassLong(hwnd, win32con.GCL_HICON)
            hicon = winapi.SendMessage(hwnd, wintraits.WM_GETICON, 0, 0)
            if hicon:
                #print('icon grabbed by WM_GETICON 0')
                pass
            if not hicon:
                hicon = winapi.SendMessage(hwnd, wintraits.WM_GETICON, 2, 0)
                if hicon:
                    #print('icon grabbed by WM_GETICON 2')
                    pass
            if not hicon:
                hicon = winapi.SendMessage(hwnd, wintraits.WM_GETICON, 1, 0)
                if hicon:
                    #print('icon grabbed by WM_GETICON 1')
                    pass
            if not hicon:
                hicon = winapi.GetClassLongPtr(hwnd, -14)
                if hicon:
                    #print('icon grabbed by GCLP_HICON')
                    pass
            if not hicon:
                hicon = winapi.GetClassLongPtr(hwnd, -34)
                if hicon:
                    #print('icon grabbed by GCLP_HICONSM')
                    pass
            
            if hicon:
                info = winapi.GetIconInfo(hicon)
                try:
                    if info[4]:  # Icon has color plane.
                        hbmp = info[4]
                        bmp = winapi.GetObject(info[4])
                        mode = "RGBA" # "RGBA"
                        width = bmp.bmWidth
                        height = bmp.bmHeight
                    else:  # Icon has no colour plane, image data stored in mask.
                        hbmp = info[3]
                        bmp = winapi.GetObject(hbmp)
                        mode = "1"
                        width = bmp.width
                        height = bmp.height // 2  # A monochrome icon contains image and XOR mask in the hbmMask.
                    b = winapi.CreateBitmapFromHandle(hbmp)
                    bits_string = b.GetBitmapBits(True)
                    img = PIL.Image.frombytes(mode, (width, height), bits_string, "raw", "BGRA")

                    #crop_box = crop_by_background(img,10)
                    #print(f'CROP-BOX: [{crop_box}] [{width}, {height}]')

                    if height>24:
                        img = img.resize((width*24//height, 24))
                    elif height<24:
                        img = img.resize((width*24//height, 24))
                    image = ImageTk.PhotoImage(img)
                    return image
                finally:
                    info[3].close()
                    info[4].close()
            else:
                print(f"Unable to grab icon for hwnd:{hwnd} [{winapi.GetWindowText(hwnd)}]")                
                img = PIL.Image.new(mode="RGB", size=(24, 24), color = (28, 28, 28))
                draw = PIL.ImageDraw.Draw(img)
                draw.ellipse((4, 4, 20, 20), fill = (128, 28, 28), outline =(78, 28, 28), width=2)
                image = ImageTk.PhotoImage(img)
                return image
        except Exception as err:
            print(err, " grabbing icon for ", hwnd, winapi.GetWindowText(hwnd))

def hwnd_dpi_rect_correct( hwnd, r ):
    dpi = winapi.GetDpiForWindow(hwnd)
    winapi.SetProcessDpiAwarenessContext

def process_create_with_desktop_process_privileges(cmdline):
    hwnd = winapi.GetShellWindow()

    # Initialize variables
    lpAttributeList = ctypes.c_void_p(0)
    dwAttributeCount = 1
    dwFlags = 0
    lpSize = ctypes.c_size_t(0)

    # First call to get the required buffer size
    winapi.InitializeProcThreadAttributeList(lpAttributeList, dwAttributeCount, dwFlags, ctypes.byref(lpSize))

    # Allocate buffer
    lpAttributeList = ctypes.cast(ctypes.create_string_buffer(lpSize.value), ctypes.POINTER(ctypes.c_void_p))

    # Second call to actually initialize the attribute list
    result = winapi.InitializeProcThreadAttributeList(lpAttributeList, dwAttributeCount, dwFlags, ctypes.byref(lpSize))
    if result:
        #if not result:
        #    raise ctypes.WinError(winapi.GetLastError())

        tid, pid = winapi.GetWindowThreadProcessId(hwnd)
        hProcess = winapi.OpenProcess(wintraits.PROCESS_CREATE_PROCESS, wintraits.FALSE, pid);   
        phProcess = ctypes.c_void_p(int(hProcess))
        
        result = winapi.UpdateProcThreadAttribute(lpAttributeList, 0, wintraits.PROC_THREAD_ATTRIBUTE_PARENT_PROCESS, ctypes.byref(phProcess), ctypes.sizeof(phProcess), None, None)
        if result:
            siex = wintypes.STARTUPINFOEX()
            siex.cb = ctypes.sizeof(wintypes.STARTUPINFOEX)
            siex.lpAttributeList = lpAttributeList
            

            pi = wintypes.PROCESS_INFORMATION()

            result = winapi.CreateProcess(None, cmdline,  None, None, wintraits.FALSE,
                                    wintraits.CREATE_NEW_CONSOLE | wintraits.EXTENDED_STARTUPINFO_PRESENT, 
                                    None, None, siex, ctypes.byref(pi))
            
            winapi.CloseHandle(pi.hThread)
            winapi.CloseHandle(pi.hProcess)
        winapi.CloseHandle(hProcess)
    return True if result!=wintraits.FALSE else False


def hwnd_system_app_user_model_id_get(hwnd):
    '''
    WARNING: winapi.CoInitialize() is required to be called once per thread before this
    '''
    ps = wintypes.PropertyStore.from_hwnd(hwnd)
    old_value =  ps[wintraits.System.AppUserModel.ID]
    return old_value 

def hwnd_system_app_user_model_id_set(hwnd, new_value):
    '''
    WARNING: winapi.CoInitialize() is required to be called once per thread before this
    '''
    ps = wintypes.PropertyStore.from_hwnd(hwnd)
    old_value =  ps[wintraits.System.AppUserModel.ID]
    ps[wintraits.System.AppUserModel.ID] = new_value
    ps.commit()
    return old_value 


def hwnd_set_layered(hwnd, color_key_rgb_triplet=None, alpha_percent=None):
    l_ex_style = winapi.GetWindowLong(hwnd, wintraits.GWL_EXSTYLE)
    l_ex_style |= wintraits.WS_EX_LAYERED
    winapi.SetWindowLong(hwnd, wintraits.GWL_EXSTYLE, l_ex_style)
    flags = 0
    color_key = 0
    alpha_byte = 0
    if color_key_rgb_triplet is not None:
        r, g, b = color_key_rgb_triplet
        color_key = winapi.RGB(r//256, g//256, b//256)
        flags |= wintraits.LWA_COLORKEY
    if alpha_percent is not None:
        alpha_byte = int(alpha_percent*255)
        flags |= wintraits.LWA_ALPHA
    if flags != 0:
        winapi.SetLayeredWindowAttributes(hwnd, color_key, alpha_byte, flags)




class tkanchor:
    from tkinter.constants import (W,E,EW,N,NW,NE,S,SE,SW,NSEW,CENTER,NS)
    '''
    W = tkcontants.W
    E = tkcontants.E
    EW = tkcontants.EW
    N = tkcontants.N
    NW = tkcontants.NW
    NE = tkcontants.NE
    S = tkcontants.S
    SE = tkcontants.SE
    SW = tkcontants.SW
    NSEW = tkcontants.NSEW
    CENTER = tkcontants.CENTER

    '''

class tkexpand:
    YES = True
    NO = False

class tkfill:
    from tkinter.constants import (X,Y,BOTH)
    '''
    X = tkcontants.X
    Y = tkcontants.Y
    BOTH = tkcontants.BOTH
    '''

class tkjustify:
    from tkinter.constants import (LEFT,CENTER,RIGHT)


class tkside:
    from tkinter.constants import (LEFT,TOP,RIGHT,BOTTOM)
    '''
    LEFT = tkcontants.LEFT
    TOP = tkcontants.TOP
    RIGHT = tkcontants.RIGHT
    BOTTOM = tkcontants.BOTTOM
    '''

class tksticky:
    from tkinter.constants import (W,E,EW,N,NW,NE,S,SE,SW,NSEW,CENTER,NS)
    '''
    W = tkcontants.W
    E = tkcontants.E
    EW = tkcontants.EW
    N = tkcontants.N
    NW = tkcontants.NW
    NE = tkcontants.NE
    S = tkcontants.S
    SE = tkcontants.SE
    SW = tkcontants.SW
    NSEW = tkcontants.NSEW
    CENTER = tkcontants.CENTER
    '''


class tkorient:
    from tkinter.constants import (HORIZONTAL,VERTICAL)


class tkevent:
    MOUSE_BUTTON_LEFT = '<Button-1>'
    MOUSE_BUTTON_MIDDLE = '<Button-2>'
    MOUSE_BUTTON_RIGHT = '<Button-3>'
    BUTTON_1 = '<Button-1>'
    BUTTON_2 = '<Button-2>'
    class mouse:
        BUTTON_LEFT = '<Button-1>'
        BUTTON_MIDDLE = '<Button-2>'
        BUTTON_RIGHT = '<Button-3>'
    class combobox:
        SELECTED = "<<ComboboxSelected>>"



    





from hPyT import *
from pywinstyles import *
from sv_ttk import *
from tkinter import font


def center(window):
    window_frame.center(window)

def center_relative(window, relative_to_window):
    window_frame.center_relative(relative_to_window, window) 





from tkinter.font import Font
from tkinter import (Variable, BooleanVar, IntVar, DoubleVar, StringVar, Menu, Listbox as ListBox, Listbox, LabelFrame, Text)
from tkinter.ttk import (Button, Entry, Frame, Label, Labelframe, LabeledScale, Radiobutton, Treeview,
                         Checkbutton, Combobox, Scrollbar, Style, Separator, Menubutton, OptionMenu)

from tkinter.ttk import (LabelFrame,
                         LabeledScale as Labeledscale, 
                         Radiobutton as RadioButton, 
                         Checkbutton as CheckButton, 
                         Checkbutton as CheckBox, 
                         Combobox as ComboBox,
                         Scrollbar as ScrollBar,
                         Menubutton as MenuButton,
                         OptionMenu as Optionmenu,
                         Treeview as TreeView)

import tkinter.ttk as ttk
import tkinter as tk

from tkinter import (Canvas, Misc as Tool)

from tkinter import Tk as AppWindow, Toplevel


class MenuBar:
    pass

class MenuCommand:
    pass

class MenuCascade:
    pass

class MenuPopup:
    pass



from . import winapi
from . import wintraits
from . import wintools
from ..smarts import smartmatch
#from ..pyrite02_tools import smartclass
#from . import tools
import sys
import types
from bitarray import bitarray
import time
import re
import math
import psutil
import win32process
import os
from sortedcontainers import SortedSet, SortedDict
from datetime import datetime

def xprint(*args, **kwargs):
    #print(*args,**kwargs,file=sys.stderr)
    pass

class P:
    def __init__(self,x=None,y=None):
        self.x = x
        self.y = y
    @staticmethod
    def from_tuple(xy):
        x, y = xy
        return P(x,y)
    def as_dict(self):
        return {k:v for k,v in self.__dict__.items() if k in ['x','y']}
    def as_tuple(self):
        return self.x, self.y    
    def moved_by(self,dx,dy):
        return P(self.x+dx,self.y+dy)

    
class R:
    def __init__(self, x=None, y=None, w=None, h=None):
        self.x, self.y, self.w, self.h = x, y, w, h        
    @staticmethod
    def from_tuple_ltrb(ltrb):
        l, t, r, b = ltrb
        return R(l,t,r-l,b-t)
    @staticmethod
    def from_tuple_xywh(xywh):
        return R(*xywh)
    def as_dict(self):
        return {k:v for k,v in self.__dict__.items() if k in ['x','y','w','h']}
    def as_tuple_pp(self):
        return (self.x,self.y),(self.x+self.w, self.y+self.h)
    def as_tuple_ltrb(self):
        return self.x,self.y, self.x+self.w, self.y+self.h
    def as_tuple_p(self, index=0):
        return self.as_tuples(index)
    def as_tuple_xywh(self):
        return self.x, self.y, self.w, self.h
    def as_tuple_wh(self):
        return self.w, self.h

    def as_tk_geometry(self):
        return f"{self.w()}x{self.h()}{self.x():+d}{self.y():+d}"
    
    def intersects_r(self, other):
        l1, t1, r1, b1 = self.as_tuple_ltrb()
        l2, t2, r2, b2 = other.as_tuple_ltrb()
        return True if l1<=r2 and r1>=l2 and t1<=b2 and b1>=t2 else False

    def contains_r(self, other):
        l1, t1, r1, b1 = self.as_tuple_ltrb()
        l2, t2, r2, b2 = other.as_tuple_ltrb()
        return True if l1<=l2 and r1>=r2 and t1<=t2 and b1>=b2 else False

    def contains_p(self, p:P):
        l1, t1, r1, b1 = self.as_tuple_ltrb()
        x, y = p.as_tuple()
        return True if l1<=x and r1>=x and t1<=y and b1>=y else False

    def stretched_by(self, from_left, from_top, from_right, from_bottom):
        l, t, r, b = self.as_tuple_ltrb()
        return R.from_tuple_ltrb(l+from_left,t+from_top,r-from_right,b-from_bottom)

    def moved_by(self, dx, dy):
        return R(self.x+dx, self.y+dy, self.w, self.h)

    def moved_to(self, x, y):
        return R(x, y, self.w, self.h)


class Mh:
    def __init__(self, hmon=None):
        self.hmon = hmon

class Wh:
    def __init__(self, hwnd=None, hosted_hwnd=None):
        self.hwnd = hwnd
        self.hosted_hwnd = hosted_hwnd
    def window_text(self):
        return winapi.GetWindowText(self.hwnd)
    def is_minimized(self):
        return True if winapi.IsIconic(self.hwnd) else False
    def is_maximized(self):
        return True if (winapi.GetWindowLong(self.hwnd,wintraits.GWL_STYLE) & wintraits.WS_MAXIMIZE) != 0 else False
    def is_normal(self):
        return True if (winapi.GetWindowLong(self.hwnd,wintraits.GWL_STYLE) & (wintraits.WS_MAXIMIZE|wintraits.WS_MINIMIZE)) == 0 else False
    def is_visible(self):
        return winapi.IsWindowVisible(self.hwnd)
    def rectangle(self):
        return winapi.GetWindowRect(self.hwnd)

class M(Mh,R):
    def __init__(self, hmon=None, x=None, y=None, w=None, h=None, name=None, number=None):
        Mh.__init__(self,hmon)
        R.__init__(self,x,y,w,h)
        self.name = name
        self.number = number
    def as_tuple(self):
        return (v for k,v in self.__dict__.items() if not k.startswith('_'))
    def as_dict(self):
        return {k:v for k,v in self.__dict__.items() if not k.startswith('_')}
    
    @staticmethod
    def from_hmon(hmon=None, number=FileNotFoundError):
        imon = winapi.GetMonitorInfo(hmon)
        l,t,r,b = imon['Work']
        return M(hmon, l, t, r-l, b-t, name=imon['Device'], number=number)
    
    @staticmethod
    def list_monitors(options=None):
        return [M.from_hmon(hmon) for hmon in wintools.hmon_list_monitors()]
    
    def fetch(options=None):
        '''
        ricarica i campi della struttura sulla base del monitor reale ammesso che esista
        ancora un monitor identificato con hmon
        '''
        pass

    def rake(options=None):
        '''
        cerca di ritrovare i monitor compatibili con gli attributi e ritorna una lista
        '''
        pass

        
class W(Wh,R):
    def __init__(self, hwnd=None, x=None, y=None, w=None, h=None, title=None
                 , pid=None, tid=None, palias=None, pname=None, pexe=None
                 , dpi=None, icon=None, wndclassname=None, hosted_hwnd=None):
        Wh.__init__(self,hwnd,hosted_hwnd)
        R.__init__(self,x,y,w,h)
        self.title = title
        self.wndclassname = wndclassname
        self.dpi = dpi
        self.pid = pid
        self.tid = tid
        self.pname = pname
        self.pexe = pexe
        self.palias = palias
        self.icon = icon
    def as_tuple(self):
        return (v for k,v in self.__dict__.items() if not k.startswith('_'))
    def as_dict(self):
        return {k:v for k,v in self.__dict__.items() if not k.startswith('_')}
    @staticmethod
    def from_w(__w__, **kwargs):
        ww = W(**__w__.as_dict())
        ww.__dict__.update(kwargs)
        return ww
    @staticmethod
    def from_tk_window(__w__, **kwargs):
        ww = W(wintools.hwnd_from_tk_window(__w__))
        ww.__dict__.update(kwargs)
        return ww

    
    @staticmethod
    def list_windows(options=None):
        exclude_handles=None, 
        exclude_states=['minimized','maximized','hidden']
        exclude_titles=['']
        limit_monitors=None        
        if options is not None:
            exclude_handles = options.get('exclude_handles',exclude_handles)   
            exclude_states = options.get('exclude_states',exclude_states)   
            exclude_titles = options.get('exclude_titles',exclude_titles)   
            limit_monitors = options.get('limit_monitors',limit_monitors)   
        def get_windows_2():
            all_windows_handle_list = []

            # The callback function that will be called for each HWND
            # all we do is append the wrapped handle
            def enum_window_proc(hwnd, lparam):
                all_windows_handle_list.append(hwnd)
                return True
            
            # loop over all the children (callback called for each)
            winapi.EnumWindows(enum_window_proc, 0)
            all_windows_list = []
            host_windows_set = SortedDict()
            hosted_windows_set = SortedSet()
            orphan_host_windows_set = SortedDict()
            for hwnd in all_windows_handle_list:
                tid, pid = winapi.GetWindowThreadProcessId(hwnd)
                try:
                    p = psutil.Process(pid)
                    if p.name()=='ApplicationFrameHost.exe':
                        try:
                            cwl = wintools.hwnd_list_child_windows(hwnd)             
                            for cw in cwl:
                                ctid, cpid = winapi.GetWindowThreadProcessId(cw)
                                if cpid != pid:
                                    hosted_windows_set.add(cw)
                                    host_windows_set[hwnd] = cw
                                    break
                            else:
                                orphan_host_windows_set[winapi.GetWindowText(hwnd)]=hwnd
                                continue
                        except Exception as e:
                            xprint('Exception in get_windows_2', e)
                    all_windows_list.append((hwnd,tid,pid,p.name())) 
                except Exception as e:
                    xprint("!!! EXCEPTION psutil.Process: ", e)

            monitors_list = M.list_monitors()
            windows_list = []
            for hwnd, tid, pid, pname in all_windows_list:
                #t_text = win32gui.GetWindowText(hwnd)
                #if t_text.find('Hyper')>=0:
                #    pass
                if hwnd in hosted_windows_set:
                    continue
                w_style = winapi.GetWindowLong(hwnd, wintraits.GWL_STYLE)
                if w_style & wintraits.WS_VISIBLE == 0:
                    continue
                w_ex_style = winapi.GetWindowLong(hwnd, wintraits.GWL_EXSTYLE)
                if w_ex_style & wintraits.WS_EX_NOACTIVATE != 0:
                    continue
                class_name = winapi.GetClassName(hwnd)
                if class_name in ['Shell_TrayWnd', 'Progman', 'WorkerW']:
                    continue
                text = winapi.GetWindowText(hwnd)
                if len(text.strip())==0:
                    continue
                if text in ['Windows Shell Experience Host', 'Windows Input Experience']:
                    continue
                if (w_style & wintraits.WS_ICONIC) == 0:
                    rect = R.from_tuple_ltrb(winapi.GetWindowRect(hwnd))
                    for monitor in monitors_list:
                        if rect.intersects_r(monitor):
                            break
                    else:
                        continue
                host_hwnd = None
                hosted_hwnd = host_windows_set.get(hwnd,None)
                if hosted_hwnd is None:
                    host_hwnd = orphan_host_windows_set.get(text,None)
                    if host_hwnd is not None:
                        hosted_hwnd = hwnd
                if host_hwnd is None:
                    host_hwnd = hwnd
                windows_list.append(W(hwnd=host_hwnd,tid=tid,pid=pid,pname=pname,hosted_hwnd=hosted_hwnd))
            return windows_list



        xprint(f'list_windows start')
        t0 = time.monotonic()
        w_all = get_windows_2() #get_windows()
        t1 = time.monotonic()
        xprint(f'get_windows end t0={t0} t1={t1} T={t1-t0}')
        w_sel = []
        exclude_visible = False
        exclude_hidden = False
        exclude_minimized = False
        exclude_maximized = False
        exclude_restored = False
        if exclude_states is not None:
            if type(exclude_states) is str:
                exclude_states = [exclude_states]    
            if type(exclude_states) is dict:
                exclude_visible = exclude_states.get('visible', False)    
                exclude_hidden = exclude_states.get('hidden', False)    
                exclude_minimized = exclude_states.get('minimized', False) or exclude_states.get('iconized', False)
                exclude_maximized = exclude_states.get('maximized', False) or exclude_states.get('zoomed', False)
                exclude_restored = exclude_states.get('restored', False) or exclude_states.get('normal', False)
            elif type(exclude_states) is list or type(exclude_states) is tuple:
                exclude_visible = 'visible' in exclude_states
                exclude_hidden = 'hidden' in exclude_states
                exclude_minimized = 'minimized' in exclude_states or 'iconized' in exclude_states
                exclude_maximized = 'maximized' in exclude_states or 'zoomed' in exclude_states
                exclude_restored = 'restored' in exclude_states or 'normal' in exclude_states

        w : W
        for z, w in enumerate(w_all):
            try:
                wt = w.window_text()
                if exclude_handles is not None:
                    if int(w.hwnd) in exclude_handles:
                        xprint(f'{w.hwnd} EXCLUDED BY HANDLE [{wt}]')
                        continue
                if limit_monitors is not None:
                    hmon = winapi.MonitorFromWindow(w.hwnd, wintraits.MONITOR_DEFAULTTONEAREST)
                    if not hmon in limit_monitors:
                        xprint(f'{w.hwnd} EXCLUDED BY MONITOR [{wt}]')
                        continue
                if exclude_titles is not None:
                    xprint(f"*** {[wt.lower()]} --- {exclude_titles}")
                    if wt.lower() in exclude_titles:
                        xprint(f'{w.hwnd} EXCLUDED BY TITLE [{wt}]')
                        continue
                if exclude_minimized and w.is_minimized():
                    xprint(f'{w.hwnd} EXCLUDED BY MINIMIZED [{wt}]')
                    continue
                if exclude_maximized and w.is_maximized():
                    xprint(f'{w.hwnd} EXCLUDED BY MAXIMIZED [{wt}]')
                    continue
                if exclude_restored and w.is_normal():
                    xprint(f'{w.hwnd} EXCLUDED BY RESTORED [{wt}]')
                    continue
                if exclude_visible and w.is_visible():
                    xprint(f'{w.hwnd} EXCLUDED BY VISIBLE [{wt}]')
                    continue
                if exclude_hidden and (not w.is_minimized()) and (not w.is_visible()):
                    xprint(f'{w.hwnd} EXCLUDED BY HIDDEN [{wt}]')
                    continue
                _x,_y,_w,_h = R.from_tuple_ltrb(w.rectangle()).as_tuple_xywh()
                w_sel.append(W.from_w(w,title=wt,x=_x,y=_y,w=_w,h=_h,ztag=f'z{z}'))
            except Exception as e:
                xprint(f'EXCEPTION in list_windows: {e}')

        t2 = time.monotonic()
        xprint(f'update list t1={t1} t2={t2} T={t2-t1}')
        xprint(f'list_windows end')
        return w_sel

    def fetch(self,options=None):
        '''
        ricarica i campi della struttura sulla base della finestra reale ammesso che esista
        ancora una finestra identificata con hwnd
        '''
        if not hasattr(self,'pcmdline'):
            p = psutil.Process(self.pid)
            self.pcmdline=p.cmdline()
            self.pexe = p.exe()
            self.pcwd = p.cwd()
        self.classname = winapi.GetClassName(self.hwnd)   
        self.model = wintools.hwnd_system_app_user_model_id_get(self.hwnd)     
        pass

    @staticmethod
    def rake_windows(options=None, **kwargs):      
        _options = dict(exclude_states=[],exclude_titles=[])
        if options is not None:
            _options = _options.update([v for v in options.get(k,None) 
                                    for k in ('exclude_handles','exclude_states','exclude_titles','limit_monitors') 
                                        if v is not None]) 
        wl = W.list_windows(_options)
        rl = []
        for z, w in enumerate(wl):
            try:
                w.fetch()
                for k,v in [(k,v) for k,v in kwargs.items() if v is not None]:
                    try:
                        vv = getattr(w,k)
                        if type(vv) is list:
                            vv = ' '.join(vv)
                        if not re.match(v,vv):
                            break
                    except Exception as e:
                        xprint(e)
                        break    
                else:
                    #w.ztag = f'z{z}'
                    rl.append(w)    
            except Exception as e:
                pass
            pass
        return rl

        '''
        cerca di ritrovare le finestre compatibili con gli attributi e ritorna una lista
        '''
        pass

    def _ensure_pname(self) -> str: 
        if self.pname is None:
            tid, pid = winapi.GetWindowThreadProcessId(self.hwnd)
            try:
                p = psutil.Process(pid)
                self.pname = p.name()                
            except:
                return '<<UNKNOWN>>'
        return self.pname

    def correct_rect(self, r:R):
        l, t, r, b = r.as_tuple_ltrb()
        if self._ensure_pname().lower() == 'code.exe':
            if winapi.GetDpiForWindow(self.hwnd)==96:
                l-=3
                r+=3
                b+=3            
        else:
            l-=7
            r+=7
            b+=7
        return R.from_tuple_ltrb((l,t,r,b))
    
    def decorrect_rect(self, r:R):
        l, t, r, b = r.as_tuple_ltrb()
        if self._ensure_pname().lower() == 'code.exe':
            if winapi.GetDpiForWindow(self.hwnd)==96:
                l+=3
                r-=3
                b-=3            
        else:
            l+=7
            r-=7
            b-=7
        return R.from_tuple_ltrb((l,t,r,b))
    
    def spread_code_guess(self, pad=None, ratio=None, options=None):
        w = self
        if winapi.IsIconic(w.hwnd) or not winapi.IsWindowVisible(w.hwnd):
            return ''
        if pad is None:
            _pad = 10
        elif type(pad) is str:
            _pad = int(pad)
        m0 = M.from_hmon(winapi.MonitorFromWindow(w.hwnd))            
        wr = w.decorrect_rect(R.from_tuple_ltrb(winapi.GetWindowRect(w.hwnd)))
        result = SpreadCode.guess(m0,_pad,wr,ratio=ratio,options=options)
        return result
    
    def spread(self, scode:str, pad=None):
        if pad is None:
            _pad = 10
        elif type(pad) is str:
            _pad = int(pad)
        w = self
        ratio = None
        if scode is not None:
            m = smartmatch()
            if m(r'^[0-9]{1,2}/[0-9]{1,2}$', scode):
                ratio = scode
                scode = None
        if scode is None:
            scode = self.spread_code_guess(ratio=ratio)
        xprint(f'W.spread spread-code={scode}')
        sc = SpreadCode(scode)
        m0 = None
        if sc.monitor is not None:
            imon = sc.monitor-1
            ml =  wintools.hmon_list_monitors()
            if imon<len(ml):
                m0 = M.from_hmon(ml[imon])
        m1 = M.from_hmon(winapi.MonitorFromWindow(w.hwnd))
        if m0 is None:
            m0 = m1
        wr = sc.spread(m0,_pad)
        _wrs = w.correct_rect(wr)
        winapi.MoveWindow(w.hwnd,*_wrs.as_tuple_xywh(),wintraits.TRUE)        
        if m0.hmon != m1.hmon:
            time.sleep(0.2)
            winapi.MoveWindow(w.hwnd,*_wrs.as_tuple_xywh(),wintraits.TRUE)
        _wrg = winapi.GetWindowRect(w.hwnd)
        _wefb = winapi.DwmGetWindowExtendedFrameBounds(w.hwnd)
        _wdpi = winapi.GetDpiForWindow(w.hwnd)
        return self.hwnd, _wrs, _wrg, _wefb, _wdpi
    
    @staticmethod
    def start(cmdline, spread=None, rake=None, options=None):
        if rake is not None:
            wl = W.rake_windows(**rake)
            if wl:
                for w in wl:
                    wintools.hwnd_restore(w.hwnd)
                    if spread is not None and options is not None and 'spread-alwais' in options:
                        w.spread(spread)
                return
        wlb:list[W] = W.list_windows(options=dict(exclude_states=[]))
        wsb = SortedSet({w.hwnd for w in wlb})
        rematch = smartmatch()
        if os.path.exists(cmdline) or rematch(r' *"? *[a-zA-Z][a-zA-Z_-]+:.*',cmdline):
            os.startfile(cmdline)
        else:
            StartupInfo = win32process.STARTUPINFO()
            StartupInfo.dwFlags = 0 #win32process.STARTF_USESHOWWINDOW
            #StartupInfo.wShowWindow = wintraits.SW_NORMAL
            flags = wintraits.NORMAL_PRIORITY_CLASS | wintraits.DETACHED_PROCESS | wintraits.CREATE_NEW_PROCESS_GROUP
            if options is not None:
                if 'new-console-window' in options:
                    flags = wintraits.NORMAL_PRIORITY_CLASS | wintraits.CREATE_NEW_CONSOLE | wintraits.CREATE_NEW_PROCESS_GROUP
                if 'no-console-window' in options:
                    flags = wintraits.NORMAL_PRIORITY_CLASS | wintraits.CREATE_NO_WINDOW | wintraits.CREATE_NEW_PROCESS_GROUP        
            hProcess, hThread, dwProcessId, dwThreadId = win32process.CreateProcess(None, cmdline, None, None, wintraits.FALSE,
                            flags, 
                            None, None, StartupInfo)
            winapi.CloseHandle(hThread)
            winapi.CloseHandle(hProcess)

        '''
        winapi.WinExec(cmdline,wintraits.SW_SHOW)
        '''
        xprint(f'{datetime.now()}w-rake-start')
        wdx = SortedDict()
        for i in range(0,10):
            time.sleep(0.2)
            wla:list[W] = W.list_windows(options=dict(exclude_states=[]))
            for w in wla:
                if w.hwnd not in wsb:
                    if w.hwnd not in wdx:
                        wdx[w.hwnd] = w                        
                        xprint(f'FOUND {w.hwnd} {w.title.encode()} {w.pname.encode()}')
                        wintools.hwnd_restore(w.hwnd)
                        if spread is not None:
                            w.spread(spread)
                            xprint(f'{datetime.now()} w-rake-spreaded to {spread}')                        
                        else:
                            xprint(f'{datetime.now()}w-rake-found')
                        return w
        xprint(f'{datetime.now()}w-rake-end')

    
'''
        _dpi = winapi.GetDpiForWindow(self.hwnd)
        if _dpi == 96:
            l-=7
            r+=7
            b+=7
        return R.from_tuple_ltrb((l,t,r,b))
'''

'''
        if self.pname.lower() == 'code.exe':
            l+=0
            r-=0
            b-=0
        else:
            l-=0
            r+=0
            b+=0
        return R.from_tuple_ltrb((l,t,r,b))
'''        
        


class SpreadCode:
    class __fields__:
        X_DIVS = 'x_divs'
        X_SPAN = 'x_span'
        X_FROM = 'x_from'
        X_AXIS = 'x_axis'
        SEPAR  = 'separ'
        Y_DIVS = 'y_divs'
        Y_SPAN = 'y_span'
        Y_FROM = 'y_from'
        Y_AXIS = 'y_axis'
        MONITOR = 'monitor'
        STRETCHES = 'stretches'

        X_FIELDS = [X_DIVS,X_SPAN,X_FROM,X_AXIS]
        Y_FIELDS = [Y_DIVS,Y_SPAN,Y_FROM,Y_AXIS] 
        FIELDS = [*X_FIELDS,
                  SEPAR,
                    *Y_FIELDS,
                    MONITOR,STRETCHES]
    
    class Stretch:
        def __init__(self, delta:int, sides:str):
            self.delta = delta
            self.sides = sides

    def __init__(self, text:str):
        self.mask : bitarray = None
        self.monitor : int = None
        self.separ : str = None
        self.x_divs : int = None
        self.x_span : int = None
        self.x_from : int = None
        self.x_axis : str = None
        self.y_divs : int = None
        self.y_span : int = None
        self.y_from : int = None
        self.y_axis : str = None
        self.stretches = []
        ms, m3 = self._parse(text)
        if ms is not None:
            self.mask = ms
            self.__dict__.update({k:v for k,v in m3.items() if k!='stretches'})
            for x in m3['stretches']:
                self.stretches.append(SpreadCode.Stretch(**x))

    @staticmethod
    def parse(text):
        return SpreadCode(text)
        
    def _parse(self, text):
        divs_s = lambda scope: fr'(?P<{scope}_divs__s>[0-9])'
        span_s = lambda scope: fr'(?P<{scope}_span__s>[0-9])'
        from_s = lambda scope: fr'(?P<{scope}_from__s>[0-9])'

        divs_l = lambda scope: fr'(?P<{scope}_divs__l>[0-9]+)'    
        span_l = lambda scope: fr'(?P<{scope}_span__l>[0-9]+)'    
        from_l = lambda scope: fr'(?P<{scope}_from__l>[0-9]+)'    
                

        stretch_d = lambda scope: fr'(?P<{scope}_stretch_delta__k>[+-][0-9]+)'
        stretch_s = lambda scope: fr'(?P<{scope}_stretch_sides__s>(?:px)?([xywhltrbXYWH]+)?)'
        stretch = lambda scope: fr'(?:{stretch_d(scope)}{stretch_s(scope)})'

        axis = lambda scope: fr'(?P<{scope}_axis__k>[xy])' 
        separ = lambda scope: fr'(?P<{scope}_separ__k>(?<![xy])[/\\]|[/\\]?)' 
        monitor = lambda scope: fr'(?:\:(?P<{scope}_monitor__k>[0-9]+))'
        part = lambda scope: \
            fr'(?:' \
                fr'{divs_s(scope)}{span_s(scope)}{from_s(scope)}' \
                fr'|' \
                fr'{divs_l(scope)}\.{span_l(scope)}\.{from_l(scope)})' \
            fr'{axis(scope)}?'   
        pattern = lambda scope: fr'^{part(f"{scope}_x")}(?:{separ(scope)}{part(f"{scope}_y")})?' \
            fr'{stretch(f"{scope}_s1")}?{stretch(f"{scope}_s2")}?' \
            fr'{stretch(f"{scope}_s3")}?{stretch(f"{scope}_s4")}?' \
            fr'{stretch(f"{scope}_s5")}?{stretch(f"{scope}_s6")}?' \
            fr'{stretch(f"{scope}_s7")}?{stretch(f"{scope}_s8")}?' \
            fr'{monitor(scope)}?$'  
        parser = re.compile(pattern('p'))
        mx = parser.match(text)
        if not mx:
            return None, None
        m1 = {k:mx.group(i) for k,i in mx.re.groupindex.items()}
        m2 = {k[2:-3]:None for k in m1} | {k[2:-3]:v for k,v in m1.items() if v is not None}
        #x = {'_'.join(reversed(k.split('_'))):v for k,v in m2.items()}
        if  (m2['x_axis'] is None and m2['y_axis'] is None and m2['separ']=='\\') or \
            (m2['x_axis'] == 'y' or m2['y_axis'] == 'x') :
            for k in [kk[2:] for kk in m2 if kk.startswith('x_')]:
                m2[f'x_{k}'], m2[f'y_{k}'] = (m2[f'y_{k}'], m2[f'x_{k}'])        
        m2.update({k:int(v) for k,v in m2.items() if v is not None and k.endswith('_offset_delta')})
        m3 = {k:v for k,v in m2.items() if k[2:11]!='_stretch_' and k[2:10]!='_spread_'} | \
                {'stretches':[{'delta':int(v), 'sides':m2[f'{k[:-6]}_sides']} for k,v in m2.items() if v is not None and k.endswith('_stretch_delta')]}
        m3.update({k:int(v) for k,v in m3.items() if v is not None and k[2:6] in('divs','from','span')})
        if m3['monitor'] is not None:
            m3['monitor'] = int(m3['monitor'])
        #m3 = m2 | {'x_axis':'x', 'y_axis':'y'}
        #mr = {k:m3[k] for k in sorted(m3)} 
        ms = bitarray([1 if v is not None else 0 for v in m3.values()])
        return ms, m3
    def bitmask(self,*keys):
        return bitarray([1 if k in keys else 0 for k in SpreadCode.__fields__.FIELDS]) if keys else self.mask
    
    @staticmethod 
    def spread_path(__lwr,__upr,__divs,__pad):
        _pad = __pad // 2
        _lwr = __lwr+_pad
        _upr = __upr-_pad
        _len = _upr - _lwr
        _slice = _len // __divs
        _extra = _len - (_slice * __divs)
        _cols = [None]*__divs
        i=1
        l = __lwr+_pad
        u = __upr-_pad
        r = u - l
        w = _slice
        e = _extra
        while r > _slice :
            if e > 0:
                c = 1
                e -=2
            else:
                c = 0
            ll = (l+_pad, l+w-_pad)
            uu = (u-w+_pad, u-_pad)
            _cols[i-1] = (*ll, ll[1]-ll[0])
            _cols[-i] = (*uu, uu[1]-uu[0])
            d = w+c
            l += d
            u -= d
            r = u - l 
            i += 1
        if _cols[i-1] is None:
            cc = (l+_pad, u-_pad)
            _cols[i-1] = (*cc, cc[1]-cc[0])
        return _cols

    
    @staticmethod
    def spread_space(__lwr,__upr,__pad,__divs,__span,__from, mode='normal'):
        _divs = __divs if __divs > 0 else 1
        _span = __span if __span > 0 and __span<=_divs else _divs            
        if __from == 0:
            if _divs % 2 == 0:
                if _span % 2 == 0:
                    _from = _divs // 2 - _span // 2
                else:       
                    _divs *= 2
                    _span *= 2
                    _from = _divs // 2 - (_span+1) // 2
            else:
                if _span % 2 == 0:
                    _divs *= 2
                    _span *= 2
                    _from = _divs // 2 - _span // 2 
                else:
                    _divs *= 2
                    _span *= 2
                    _from = _divs // 2 - (_span+1) // 2 
        else:
            _from = __from - 1

        dd = _divs - (_from+_span)
        if dd < 0:
            dd = -dd
            _from -= dd
            if _from < 0:
                dd = -_from
                _from = 0
                _span -= (dd)
        if _divs <= 0:
            _divs = 1 
        if _from < 0 or _from >= _divs:
            _from = 0
        if _span < 1 or _span > _divs:
            _span = _divs
        '''
        _pad = __pad // 2
        _lwr = __lwr+_pad
        _upr = __upr-_pad
        _len = _upr - _lwr
        _slice = _len // _divs
        _extra = _len - (_slice * _divs)
        _cols = [None]*_divs
        i=1
        l = __lwr+_pad
        u = __upr-_pad
        r = u - l
        w = _slice
        e = _extra
        while r > _slice :
            if e > 0:
                c = 1
                e -=2
            else:
                c = 0
            ll = (l+_pad, l+w-_pad)
            uu = (u-w+_pad, u-_pad)
            _cols[i-1] = (*ll, ll[1]-ll[0])
            _cols[-i] = (*uu, uu[1]-uu[0])
            d = w+c
            l += d
            u -= d
            r = u - l 
            i += 1
        if _cols[i-1] is None:
            cc = (l+_pad, u-_pad)
            _cols[i-1] = (*cc, cc[1]-cc[0])
        '''
        _cols = SpreadCode.spread_path(__lwr,__upr,_divs,__pad)
        lll = []
        cc = __lwr
        for ll, uu, xx in _cols:
            lll.append(ll-cc)
            cc = uu
        lll.append(__upr-cc)
        _cols.append(('*', *lll))
        if mode=='mirror':
            _from = _divs - _from - _span
        _res = (_cols[_from][0], _cols[_from+_span-1][1]) 
        xprint(f'spread_space({__lwr},{__upr},{__pad},{__divs},{__span},{__from}): [{_cols}] -> {_res}')
        return _res

    def spread(self, ws_rect:R, pad: int=10) -> R:
        wsl,wst,wsr,wsb = ws_rect.as_tuple_ltrb()
        rsl, rst, rsr, rsb = (wsl,wst,wsr,wsb)        
        rsm : bitarray = self.bitmask() 
        xm : bitarray = self.bitmask(*SpreadCode.__fields__.X_FIELDS) 
        ym : bitarray = self.bitmask(*SpreadCode.__fields__.Y_FIELDS) 

        x_from, x_span, x_divs = self.x_from, self.x_span, self.x_divs
        y_from, y_span, y_divs = self.y_from, self.y_span, self.y_divs
        for s in self.stretches:
            ss = s.sides
            dd = s.delta
            if ss.startswith('px'):
                ss = ss[2:]
            if ss!='':
                for c in ss:
                    if c=='X':
                        x_from += dd
                    elif c=='Y':
                        y_from += dd
                    elif c=='W':
                        x_span += dd
                    elif c=='H':
                        y_span += dd

        if (rsm & xm).count(1):         
            rsl, rsr = SpreadCode.spread_space(rsl, rsr, pad, x_divs, x_span, x_from)   
        else:
            rsl, rsr = (rsl+pad, rsr-pad)
        if (rsm & ym).count(1):            
            rst, rsb = SpreadCode.spread_space(rst, rsb, pad, y_divs, y_span, y_from, mode='mirror')
        else:
            rst, rsb = (rst+pad, rsb-pad)
        for s in self.stretches:
            ss = s.sides
            dd = s.delta
            if ss.startswith('px'):
                ss = ss[2:]
            if ss!='':
                for c in ss:
                    if c=='l':
                        rsl+=dd
                    elif c=='t':
                        rst-=dd
                    elif c=='r':
                        rsr+=dd
                    elif c=='b':
                        rsb-=dd
                    elif c=='x':
                        rsl+=dd
                        rsr+=dd
                    elif c=='y':
                        rst-=dd
                        rsb-=dd
                    elif c=='w':
                        rsl-=dd
                        rsr+=dd
                    elif c=='h':
                        rst-=dd
                        rsb+=dd
            else:
                rsl-=dd
                rst-=dd
                rsr+=dd
                rsb+=dd

        #return rsl-4, rst, rsr+4, rsb+4
        return R.from_tuple_ltrb((rsl, rst, rsr, rsb))
    

    @staticmethod
    def guess1(ws_rect:R, pad, w_rect:R):
        wsl,wst,wsr,wsb = ws_rect.as_tuple_ltrb()
        wl, wt, wr, wb = w_rect.as_tuple_ltrb()

        dd = 1000000
        for cc in range(1,10):
            _cols = SpreadCode.spread_path(wsl,wsr,cc,pad)
            for rc in range(1,10):
                _rows = SpreadCode.spread_path(wst,wsb,rc,pad)
                d0 = 1000000
                d1 = 1000000
                for i in range(0,cc):
                    for j in range(0,rc):
                        d0x = wl-_cols[i][0]
                        d0y = wt-_rows[j][0]
                        d0d = math.sqrt(d0x**2+d0y**2)
                        if d0d < d0:
                            d0 = d0d
                            d0i = i
                            d0j = j
                        d1x = wr-_cols[i][1]
                        d1y = wb-_rows[j][1]
                        d1d = math.sqrt(d1x**2+d1y**2)
                        if d1d < d1:
                            d1 = d1d
                            d1i = i
                            d1j = j
                d3 = d1 + d0
                if d3 < dd:
                    dd = d3
                    d_cols = cc
                    d_rows = rc
                    d0from = d0i+1
                    d0span = d1i-d0i+1
                    d1from = d0j+1
                    d1span = d1j-d1j+1
                    if dd<0.1:
                        break
                if dd<0.1:
                    break
            if dd<0.1:
                break
        
        if [x for x in (d_cols,d0span,d0from,d_rows,d1span,d1from) if x >=10]:
            scode = f'{d_cols}.{d0span}.{d0from}/{d_rows}.{d1span}.{d1from}'
        else:
            scode = f'{d_cols}{d0span}{d0from}/{d_rows}{d1span}{d1from}'
        return scode

    @staticmethod
    def area_of_quadrilateral(vertices):
        """
        Calculate the area of a quadrilateral given its vertices.
        
        Parameters:
        vertices (list of tuples): A list of four tuples, each containing the (x, y) coordinates of a vertex.
        
        Returns:
        float: The area of the quadrilateral.
        """
        #if len(vertices) != 4:
        #    raise ValueError("There must be exactly four vertices.")
        #v = [(x,-y) for x,y in vertices]
        v = vertices
        x1, y1 = v[0]
        x2, y2 = v[1]
        x3, y3 = v[2]
        x4, y4 = v[3]
        
        # Using the Shoelace formula
        #area = 0.5 * abs(x1*y2 + x2*y3 + x3*y4 + x4*y1 - y1*x2 - y2*x3 - y3*x4 - y4*x1)
        area = 0.5 * abs(x1*y2 + x2*y3 + x3*y4 + x4*y1 - y1*x2 - y2*x3 - y3*x4 - y4*x1)
        
        return area



    @staticmethod
    def guess(ws_rect:R, pad, w_rect:R, ratio=None, options=None):
        wsl,wst,wsr,wsb = ws_rect.as_tuple_ltrb()
        wl, wt, wr, wb = w_rect.as_tuple_ltrb()
        t0 = ((wb-wt)/(wr-wl))
        dd = 1000000000

        if ratio is not None:
            scols, srows = ratio.split('/')
            lcols, lrows = (int(scols), int(srows))
            ucols, urows = (lcols+1, lrows+1)
        else:
            lcols, lrows = (1,1)
            ucols, urows = (10,10)
        for cc in range(lcols,ucols):
            _cols = SpreadCode.spread_path(wsl,wsr,cc,pad)
            for rc in range(lrows,urows):
                _rows = SpreadCode.spread_path(wst,wsb,rc,pad)
                for x1 in range(0,cc):
                    for y1 in range(0,rc):
                        for x2 in range(x1,cc):
                            for y2 in range(y1,rc):
                                l = _cols[x1][0]
                                t = _rows[y1][0]
                                r = _cols[x2][1]
                                b = _rows[y2][1]
                                t1 = math.sqrt( (t-wt)**2 + (l-wl)**2 )
                                t2 = math.sqrt( (b-wb)**2 + (r-wr)**2 )
                                ax = (t1+t2)/2
                                if ax < dd:
                                    dd = ax
                                    rows = _rows
                                    cols = _cols
                                    dc = cc
                                    dr = rc
                                    dx1 = x1
                                    dy1 = y1
                                    dx2 = x2
                                    dy2 = y2
                                    ddt = t
                                    ddl = l
                                    ddr = r
                                    ddb = b
                    
        # dr-oj perchè la y la volgio dare come le ordinate, che crescono salendo == mirrored di sopra
        for x in (dc,dx2-dx1+1,dx1+1,dr,dy2-dy1+1,dr-dy2):
            if x>=10:
                scode = f'{dc}.{dx2-dx1+1}.{dx1+1}/{dr}.{dy2-dy1+1}.{dr-dy2}'
                break
        else:        
            scode = f'{dc}{dx2-dx1+1}{dx1+1}/{dr}{dy2-dy1+1}{dr-dy2}'
        if options is not None:
            if 'complete' in options:
                deltas = [(wl-ddl,'l'), (wt-ddt,'t'), (wr-ddr,'r'), (wb-ddb,'b')]
                d1=[]
                for _a, _b in deltas:
                    if _a != 0:
                        for _i, (_c, _d), in enumerate(d1):
                            if _c==_a:
                                d1[_i]=(_c, _d+_b)
                                break
                            elif _b=='r' and 'l' in _d and _a==-_c:
                                d1[_i]=(_a if _a<0 else _c, _d.replace('l','w'))
                                break
                            elif _b=='b' and 't' in _d and _a==-_c:
                                d1[_i]=(_a if _a<0 else _c, _d.replace('t','h'))
                                break
                        else:
                            d1.append((_a,_b))
                for _a, _b in d1:
                    scode += f'{_a:+d}{_b}'
        if isinstance(ws_rect,M):
            m0 = ws_rect
            ml = M.list_monitors()
            for i, m1 in enumerate(ml):
                if m1.hmon == m0.hmon:
                    scode += f':{i+1}'
                    break
        if options is not None:
            if 'debug' in options:
                return scode, (dd,(wl,wt,wr,wb),(ddl,ddt,ddr,ddb))
        return scode

                        





        


    
    








from . import gui06x2_tktraits as tktraits
from . import gui08x2_tktypes as tktypes
from . import gui07x2_tktools as tools
from . import gui05x1_tkext as tkext
from . import gui16x1_winapi as winapi
from . import gui14x1_wintraits as wintraits
from .gui04x1_tkliterals import *
from ..helpers import signature
import sys

#from pycore import smartscope

#import types

class context:
    app : tktypes.AppWindow = None

    @staticmethod
    def init(title=None, visible=True, iconized=False, w=800, h=600, window_class=None, **kwargs):
        if context.app is None:
            winapi.SetProcessDpiAwarenessContext(wintraits.DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2)
            if window_class is None:
                context.app = tktypes.AppWindow(**kwargs)
                #context.app.withdraw()
                #gui.tktools.set_theme('dark')
                #tools.set_theme("dark")            
                context.app.title(title)
                #context.app.geometry(f'{w}x{h}')
                #tktools.window_frame.center(app)
                #hidden=False
                #if iconized:
                #    context.app.after(300,lambda window=context.app:window.state('iconic') if winapi.GetParent(winapi.GetParent(window.winfo_id()))==0 else None)
                #    hidden=True
                #if not visible:
                #    hidden=True
                #if not hidden:
                #    context.app.after(300,lambda window=context.app:window.state('normal') if winapi.GetParent(winapi.GetParent(window.winfo_id()))==0 else None)
                
            else:
                context.app = window_class(**kwargs)
        return context.app

def run():
    context.app.mainloop()

def after(*args,**kwargs):
    return context.app.after(*args,**kwargs)


def event_generate(*args,**kwargs):
    return context.app.event_generate(*args,**kwargs)

init = context.init




from .l01x0_common_layers import Layer
#from .. import gui
from ...helpers import signature
from ...gui import gui04x1_tkliterals as tkliterals
from ...gui import gui06x2_tktraits as tktraits
from ...gui import gui07x2_tktools as tktools
from ...gui import gui08x2_tktypes as tktypes
from ...gui import tkfonts as tkfonts
from ...gui.gui04x1_tkliterals import *
from ...gui import init, run, after

import sys 

class tklayout:
    class pack(signature): pass
    class grid(signature): pass
    class place(signature): pass

class tkinit(signature):
    default = ('', (), {})

class tkconfigure(signature):
    pass

class tkstyle(signature):
    default = ('', (), {})
    def configure(self, *args):
        style_name = self[1][0]
        for class_name in args:
            if not style_name.endswith(f'.{class_name}'):
                complete_name = f'{style_name}.{class_name}'
            else:
                complete_name = style_name
            tktypes.Style().configure(complete_name, *self[1][1:], **self[2])

class TkGuiLayer(Layer):
    def item_build(self, builder):
        scope = builder.scope
        this = None
        if type(builder.type) is type:
            if issubclass(builder.type,tktypes.Tool):  
                init_signature = builder.kwargs.get('init', None)
                method, args, kwargs = init_signature if init_signature is not None else tkinit.default
                if scope.__joint__.base.__joint__.tool is not None:
                    master = getattr(scope.__joint__.base.__joint__.tool,'interior_frame',None)
                    if master is None:
                        master = getattr(scope.__joint__.base.__joint__.tool,'client_frame',None)
                    if master is None:
                        master = scope.__joint__.base.__joint__.tool
                    this = builder.type(master, *args, **kwargs)
                else:
                    if issubclass(builder.type, tktypes.AppWindow): 
                        theme_signature = builder.kwargs.get('theme', None)
                        if theme_signature:
                            del builder.kwargs['theme']
                        this = init(*args, **kwargs)
                        if theme_signature:
                            if type(theme_signature) is str:
                                tktools.set_theme(theme_signature)
                    else:
                        this = builder.type(*args, **kwargs)
            else:
                init_signature = builder.kwargs.get('init', None)
                if init_signature:
                    method, args, kwargs = init_signature 
                    this = builder.type(*args, **args)
                else:
                    this = builder.type()
        else:
            if type(builder.item) is type:
                this = builder.item()     
        if this is None:
            this = builder.item
        if this is not None:
            if isinstance(this,tktypes.Tool):
                scope.__joint__.tool = this
            else:
                scope.__joint__.mass = this  
        return this

    def item_stage(self, builder):
        if type(builder.type) is type:
            if issubclass(builder.type,tktypes.Tool):
                this = builder.this
                style_signature = builder.kwargs.get('style', None)
                if style_signature is not None:                    
                    class_name = this.winfo_class()
                    if type(style_signature) is str:
                        style_name = style_signature
                        if style_name not in ('','.'):
                            if not style_name.endswith(f'.{class_name}'):
                                style_name = f'{style_name}.{class_name}'
                        else:
                            style_name = class_name
                    else:
                        method, args, kwargs = style_signature
                        if args:
                            style_name = args[0]
                            args = args[1:]
                        else:
                            style_name = this.winfo_name()
                        if not style_name.endswith(f'.{class_name}'):
                            style_name = f'{style_name}.{class_name}'
                        if args or kwargs:
                            tktypes.Style().configure(style_name,*args,**kwargs)
                    this.configure(style=style_name)

                cfg_sig = builder.kwargs.get('configure',None)
                if cfg_sig is None:
                    cfg_sig = builder.kwargs.get('config',None)
                if cfg_sig is not None:
                    method, args, kwargs = cfg_sig
                    this.configure(*args,**kwargs)

                self.item_bind(builder)
                '''
                bind_spec = builder.kwargs.get('bindings',None)
                if bind_spec is None:
                    bind_spec = builder.kwargs.get('bind',None)
                if bind_spec is not None:
                    if bind_spec[0]=='<':
                        builder.this.bind(bind_spec,lambda *a,scope=builder.scope,**b:builder.wrapper(scope,*a,**b))
                    else:                        
                        builder.this.configure(**{bind_spec:builder.wrapper})
                '''
                layout_signature = builder.kwargs.get('layout',None)
                if builder.context.stack:
                    if layout_signature is None:
                        layout_signature = builder.context.trail.get('layout',None)
                    else:
                        builder.context.trail['layout'] = layout_signature
                if layout_signature is not None:
                    method, args, kwargs = layout_signature
                    method_instance = getattr(builder.this,method)
                    method_instance(*args, **kwargs)
                    #print(layout)

    def item_bind(self, builder):
        scope = builder.scope
        joint = scope.__joint__
        this = joint.this        
        if isinstance(this,tktypes.Tool):
            bind_spec = builder.kwargs.get('bindings',None)
            if bind_spec is None:
                bind_spec = builder.kwargs.get('bind',None)
            if bind_spec is not None:
                if bind_spec[0]=='<':
                    this.bind(bind_spec,lambda *a,scope=scope,**b:builder.wrapper(scope,*a,**b))
                elif bind_spec[0]=='[' and bind_spec[-1]==']':
                    this.protocol(bind_spec[1:-1], lambda scope=scope:builder.wrapper(scope))
                else:                        
                    this.configure(**{bind_spec:builder.wrapper})





#from typing import Any
#from builtins import type as builtin_type
#import itertools
import types
import functools
#from functools import partial
from . import layers
#from __future__ import annotations
#import logging
# Example of logging a debug message
#logging.debug("This is a debug message.")



# element
# ogni funzione o classe decorata con @element diventa un simbolo globale nella gerarchia di strutturazione dell'applicazione.
# Un element è un nodo di gestione per un compito specifico
# - può essere un elemento di interfaccia utente, come un widget o una window
# - può essere un elemento di gestione dati, tipo una tablle, un record.

# scope
# - [0]: root scope
# - [1]: scope appena sotto root percorrendo lo scope-bhanch all'indietro
# - [-1]: base scope
# - ['<path>']: scope at path, in teoria potremmo gestire anche le stringhe quì ed usare anche un approccio tipo pathname uri
# - base: base scope
# - mass: elemento più vicino, precorrendo lo scope-branch all'indietro, che è di categoria "data" (table,record,ecc)
# - tool: elemento più vicino, precorrendo lo scope-branch all'indietro, che è di categoria "tool" (widget,window,ecc)


# @element decorator
#def element(scope=scope_root, type=None, *args, **kwargs):
#    return lambda item, scope=scope, type=type, args=args, kwargs=kwargs: element_factory(scope,type,item,*args,**kwargs)

# penso che se ElementDecorator.__call__() ritorna una istanza di una classe fatta come sotto, 
# gestendo opportunamente __call__, __enter__ e __exit__,
# si dovrebbe poter usare il simbolo "element" si come decoratore che come context-manager

def dprint(*args,**kwargs):
    pass
class Wrapper:
    pass
class Joint:
    def Wrapper(_scope, _callee, *_args, **_kwargs) -> type:
        class WrapperType(Wrapper):
            base_scope = _scope
            base_kwargs = dict(_kwargs)
            base_args = list(_args)
            base_callee = staticmethod(_callee)
            ismethod = isinstance(_callee,types.MethodType)
            def free(self):
                if type(self) is not type:
                    for k in self.kwargs:
                        x = self.__dict__[k] 
                        self.__dict__[k] = f'id(0x{id(x):X})'
                    self.kwargs.clear()
                    self.callee = f'id(0x{id(self.callee):X})'
                    self.scope = f'id(0x{id(self.scope):X})'
                self.base_args.clear()
                self.base_kwargs.clear()
                self.base_callee = f'id(0x{id(self.base_callee):X})'
                self.base_scope = f'id(0x{id(self.base_scope):X})'
                
            def scoped(self):
                self.scope.__joint__.add_referrer('wrapper',self)
                return self
            
            def  __init__(self, **_kwargs):
                functools.update_wrapper(self,self.base_callee)
                self.scope = self.base_scope
                self.callee = self.base_callee
                self.kwargs = dict(self.base_kwargs)
                self.kwargs.update(_kwargs)
                self.__dict__.update(self.kwargs)
            def __get__(self, obj, owner):
                if self.ismethod:
                    return types.MethodType(self,obj)
                else:
                    return self
            def __call__(self, *_args, **_kwargs):
                return self.callee(self,*_args,**_kwargs)
            def select_scope(self, *scopes):
                for scope in scopes:
                    if scope is not None:
                        return scope
                else:
                    return self.scope
            def divert_scope(self, mold, molded):
                if self.scope is mold:
                    return self.select_scope(molded)
                else:
                    return self.scope
        _scope.__joint__.add_referrer('Wrapper',WrapperType)
        return WrapperType
    
    def __init__(self, name=None, pool=None, base=None, mass=None, tool=None, 
                 call=None, call2=None, this=None, mold=None, layer=None, 
                 host=None, tag=None, overrides=None):
        self.name = name
        self.pool = pool
        self.base : Scope = base
        self.mass = mass if mass is not None else base.__joint__.mass if base is not None else None
        self.tool = tool if tool is not None else base.__joint__.tool if base is not None else None
        self.call2 = call2
        self.call = call
        self.this = this
        self.mold = mold
        if layer is None:
            if mold is not None:
                layer = mold.__joint__.layer
        if layer is None:
            if base is not None:
                layer = base.__joint__.layer
        if layer is None:
            layer = layers.common.CommonLayer()
        self.layer = layer 
        self.host = host
        self.tag = tag
        self.overrides = overrides
        self.routines = {}
        self.referrers = []
        #self.features = {}
        self.counter = 0
        self.bond = None

    def add_referrer(self,role,scope=None,drop=None):
        self.referrers.append((role,scope,drop))

    def __del__(self):
        dprint(f"__del__: {self}")
        #logging.debug(f"__del__: {self}")
        pass

    def __repr__(self):
        return f'Joint tag:"{self.tag}" at:0x{id(self):X}({id(self)})'

    def free(self, scope=None):
        #self.features = f"id(0x{id(self.features):X})"

        for _role, _scope, _drop in reversed(self.referrers):
            if _role in ('element','feature','cluster','section'):
                _scope.__joint__.free(_scope)
            elif _role in ('binding','routine','attribute'):
                __scope, __name = _scope
                ref = getattr(__scope,__name,None)
                if ref is not None:
                    setattr(__scope,__name,f'id(0x{id(ref):X})')
            elif _role=='base':
                __scope, __name = self.base, self.name
                ref = getattr(__scope,__name,None)
                if ref is not None:
                    setattr(__scope,__name,f'id(0x{id(ref):X})')

        for _role, _scope, _drop in reversed(self.referrers):
            if _role == 'Wrapper':
                _scope.free(_scope)
            elif _role == 'wrapper':
                _scope.free()

        self.referrers.clear()
        #for x, (y,z) in reversed(self.routines.items()):
        #    for w in z:
        #        w.free(w)
        #    y.free()
        self.routines.clear()
        #self.routines = f"id(0x{id(self.routines):X})"
        if self.call2 is not None:
            self.call2 = f"id(0x{id(self.call2):X})"
        if self.call is not None:
            self.call = f"id(0x{id(self.call):X})"
        if self.mold is not None:
            self.mold = f"id(0x{id(self.mold):X})"
        if self.host is not None:
            self.host = f"id(0x{id(self.host):X})"
        if self.layer is not None:
            self.layer = f"id(0x{id(self.layer):X})"
        if self.base is not None:
            self.base = f"id(0x{id(self.base):X})"
        if self.tool is not None:
            self.tool = f"id(0x{id(self.tool):X})"
        if self.mass is not None:
            self.mass = f"id(0x{id(self.mass):X})"
        if self.this is not None:
            self.this = f"id(0x{id(self.this):X})"
        #if self.pool is not None:
        #    self.pool = f"id(0x{id(self.pool):X})"



    def routine_wrap(self, scope, host, name, item, aggregation=None, **kwargs):
        final_name = name
        final_wrapper = None
        final_host = None
        r_list = None        
        if host is None:
            if scope.__joint__.host is not None:
                final_host = scope.__joint__.host
                final_wrapper, r_list = scope.__joint__.host.__joint__.routines.get(final_name,(None,None))
        else:
            if isinstance(host, types.FunctionType):
                final_wrapper, _list = getattr(host,'__sequence__',(None,None))
            elif isinstance(host, types.MethodType) and isinstance(host.__self__,Scope):
                #host.__func__ is host.__self__.__joint__.routines['free'][0]  -> True
                r_dict = host.__self__.__joint__.routines
                for n, (r, l) in r_dict.items():
                    if r is host.__func__:
                        break
                else:
                    raise Exception('routine_wrap failed')    
                final_name = n
                final_wrapper = r
                r_list = l
                final_host = host.__self__
                pass
            elif isinstance(host,Scope):         
                final_host = host       
                final_wrapper, _list = getattr(host,'__sequence__',(None,None))
            else:
                raise Exception('routine_wrap failed')
        class RoutineWrapper(Joint.Wrapper(scope,item)):
            def __call__(self, _scope=None, *args, **kwargs):
                with self.select_scope(_scope) as scope:
                    return self.callee(scope,*args,**kwargs)        
        if final_wrapper and r_list:
            r_list.append(RoutineWrapper)
            class FinalWrapper(Joint.Wrapper(scope, item, host=final_host, accessor=final_wrapper)):
                def __call__(self, _scope=None, *args, **kwargs):
                    return self.accessor(self.host,*args,**kwargs)
            final_wrapper = FinalWrapper().scoped()
        else:
            if not aggregation:
                aggregation = 'ordered'
            if aggregation=='reversed':
                class FinalWrapper(Joint.Wrapper(scope,item,final_name=final_name)):
                    def __call__(self, _scope=None, *args, **kwargs):
                        with self.select_scope(_scope) as enclosing_scope:
                            _, _list = enclosing_scope.__joint__.routines[self.final_name]
                            for Wrapper in reversed(_list):
                                wrapper = Wrapper(detached=True, enclosing_scope=enclosing_scope)
                                with wrapper.scope as scope:
                                    wrapper(scope,*args,**kwargs)                    
            else:
                class FinalWrapper(Joint.Wrapper(scope,item,final_name=final_name)):
                    def __call__(self, _scope=None, *args, **kwargs):
                        with self.select_scope(_scope) as enclosing_scope:
                            _, _list = enclosing_scope.__joint__.routines[self.final_name]
                            for Wrapper in _list:
                                wrapper = Wrapper(detached=True, enclosing_scope=enclosing_scope)
                                with wrapper.scope as scope:
                                    result = wrapper(scope,*args,**kwargs)                    
                                    if result is not None:
                                        return result
            routine_wrapper = FinalWrapper().scoped()
            self.routines[final_name] = (routine_wrapper, [RoutineWrapper])
            final_wrapper = routine_wrapper        
        setattr(scope, item.__name__, types.MethodType(final_wrapper, scope))
        scope.__joint__.add_referrer('routine',(scope, item.__name__))
        return final_wrapper
        
    def feature_wrap(self, scope, name, item):
        host = self.host
        helper = item
        final_name = name
        final_host = host
        '''
        qui bisogna ragionare sul fatto che si abbia o meno un host altrimenti non funziona un piffero
        dato che ora è sempre una funzione 
        '''
        #raise Exception('Leggi il commento sopra')
        if host is None:
            class FeatureWrapper(Joint.Wraper(scope,item)):
                def __call__(self,_scope=None,*args,**kwargs):
                    with self.select_scope(_scope) as scope:
                        self.callee(scope,*args,**kwargs)
            self.routines[final_name] = (FeatureWrapper(), [FeatureWrapper])
            wrapper = final_wrapper
            target = item
        else:
            final_wrapper, _list = final_host.__joint__.routines[final_host.__name__]
            class FeatureWrapper(Joint.Wrapper(scope,item,molded=None)):
                def __call__(self, __scope=None, *args, **kwargs):                    
                    if self.molded is not None:                      
                        _scope = self.scope  
                        _joint = _scope.__joint__
                        _joint.counter += 1
                        joint = Joint(name=_joint.name, pool=_joint.pool,base=_joint.base,
                                      mass=_joint.mass,call=_joint.call,call2=_joint.call2,this=_joint.this,
                                      overrides=_joint.overrides, host=self.molded, mold=_scope, tag=f'{_joint.tag}#{_joint.counter}')
                        scope = type(_scope)(joint)
                        with scope:
                            res = self.callee(scope,*args,**kwargs)                
                            if res is not None:
                                pass 
            _list.append(FeatureWrapper)
            wrapper = final_wrapper
            target = item
        def simple_wrapper(scope, *args, **kwargs):
            with scope:
                helper(scope,*args,**kwargs)
        helper.wrapper = simple_wrapper
        self.name = final_name        
        scope.__joint__.base.__joint__.add_referrer('feature',scope)
        scope.__joint__.add_referrer('base')
        return target, wrapper, helper, final_name

    def element_wrap(self, scope, name, item):
        final_name = name
        if isinstance(item, types.FunctionType):
            class ElementWrapperAccessor(Joint.Wrapper(scope,item,final_name=final_name)):
                def __call__(self,_scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as enclosing_scope:
                        _, _list = enclosing_scope.__joint__.routines[self.final_name]
                        for Wrapper in _list:
                            wrapper = Wrapper(enclosing_scope=enclosing_scope)
                            with wrapper.scope as scope:
                                wrapper(scope,*args,**kwargs)
            element_wrapper = ElementWrapperAccessor()
            class ElementWrapper(Joint.Wrapper(scope,item)):
                def __call__(self, _scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as scope:
                        self.callee(scope,*args,**kwargs)
            self.routines[final_name] = (element_wrapper, [ElementWrapper])
            helper = item
            wrapper = element_wrapper
            target = wrapper
        else:
            raise Exception('UNIMPLEMENTED')
        class SimpleWrapper(Joint.Wrapper(scope,helper)):
            def __call__(self, _scope=None, *args, **kwargs):
                with self.select_scope(_scope) as scope:
                    self.callee(scope,*args,**kwargs)
        simple_wrapper = SimpleWrapper().scoped()
        helper.wrapper = simple_wrapper
        self.name = final_name
        scope.__joint__.base.__joint__.add_referrer('element',scope)
        scope.__joint__.add_referrer('base')
        return target, wrapper, helper, final_name
    
    def section_wrap(self, scope, name, item):
        final_name = name
        if isinstance(item, types.FunctionType):
            class SectionWrapper(Joint.Wrapper(scope, item)):
                def __call__(self, _scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as scope:
                        self.callee(scope,*args,**kwargs)
            class SectionAccessor(Joint.Wrapper(scope, item, final_name=final_name)):
                def __call__(self, _scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as enclosing_scope:
                        _, _list = enclosing_scope.__joint__.routines[self.final_name]
                        for Wrapper in _list:
                            wrapper = Wrapper(enclosing_scope=enclosing_scope)
                            with wrapper.scope as scope:                                
                                wrapper(scope,*args,**kwargs)
            section_wrapper = SectionAccessor()
            self.routines[final_name] = (section_wrapper, [SectionWrapper])
            helper = item
            wrapper = section_wrapper
            target = wrapper
        else:
            raise Exception('UNIMPLEMENTED')
        class SimpleWrapper(Joint.Wrapper(scope,helper)):
            def __init__(self,*args,**kwargs):
                super().__init__(*args,**kwargs)
                pass
            def __call__(self, _scope=None, *args, **kwargs):
                with self.select_scope(_scope) as scope:
                    self.callee(scope,*args,**kwargs)
        simple_wrapper = SimpleWrapper().scoped()
        helper.wrapper = simple_wrapper
        self.name = final_name
        scope.__joint__.base.__joint__.add_referrer('section',scope)
        scope.__joint__.add_referrer('base')
        return target, wrapper, helper, final_name

    def cluster_wrap(self, scope, name, item):
        final_name = name
        if isinstance(item, types.FunctionType):
            class ClusterWrapper(Joint.Wrapper(scope,item)):
                def __call__(self, _scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as scope:
                        self.callee(scope,*args,**kwargs)
            class ClusterAccessor(Joint.Wrapper(scope,item,final_name=final_name)):
                def __call__(self, _scope=None, *args, **kwargs):
                    with self.select_scope(_scope) as mold:
                        _joint = mold.__joint__
                        _joint.counter += 1
                        joint = Joint(name=_joint.name, pool=_joint.pool,base=_joint.base,mass=_joint.mass,
                                    call=_joint.call,this=_joint.this,host=_joint.host,mold=mold,
                                    tag=f'{_joint.tag}#{_joint.counter}') 
                        molded = type(mold)(joint)
                        _, _list = mold.__joint__.routines[self.final_name]
                        for Wrapper in _list:
                            wrapper = Wrapper(molded=molded)                                        
                            with wrapper.divert_scope(mold, molded) as scope:
                                wrapper(scope,*args,**kwargs)
            cluster_wrapper = ClusterAccessor()
            self.routines[final_name] = (cluster_wrapper, [ClusterWrapper])
            helper = item
            wrapper = cluster_wrapper
            target = wrapper
        else:
            raise Exception('UNIMPLEMENTED')
        class SimpleWrapper(Joint.Wrapper(scope,helper)):
            def __call__(self, _scope=None, *args, **kwargs):
                with self.select_scope(_scope) as scope:
                    self.callee(scope,*args,**kwargs)
        simple_wrapper = SimpleWrapper().scoped()
        helper.wrapper = simple_wrapper
        self.name = final_name
        scope.__joint__.base.__joint__.add_referrer('cluster',scope)
        scope.__joint__.add_referrer('base')
        return target, wrapper, helper, final_name    
    
    def binding_wrap(self, scope, name, item):
        final_name = name
        helper = None
        wrapper = None
        target = item
        self.name = final_name
        scope.__joint__.add_referrer('binding',(scope,self.name))
        return target, wrapper, helper, final_name
    
    def realize(self, scope, *args, **kwargs):
        if self.bond:
            return self.bond(scope,*args,**kwargs)
    #def __del__(self):
    #    print(f'==> Joint.__del__ {self.tag} id={id(self):X} ')
      
class Scope:
    def __init__(self, joint:Joint):
        self.__joint__:Joint = joint
    def __enter__(self):        
        self.__joint__.pool.select(self)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__joint__.pool.restore()
    def __call__(self, *args, **kwargs):
        return self.__joint__.call(*args, **kwargs)
    def __getattr__(self, name):
        return getattr(self.__joint__.this,name)
    def __setattr__(self, name, value):
        inner = self.__dict__.get('__joint__',None)
        if inner:
            inner = inner.__dict__.get('this', None)
            if inner:
                if hasattr(inner,name):
                    return setattr(inner,name,value)
        return object.__setattr__(self,name,value)
    def __getitem__(self, key):
        return  self.__joint__.this.__getitem__(key)
    def __setitem__(self, key, value):
        return  self.__joint__.this.__setitem__(key, value)
    def __iter__(self):
        return self.__joint__.this.__iter__()
    def __del__(self):
        dprint(f"__del__: Scope of {self.__joint__}")
        #logging.debug(f"__del__: Scope of {self.__joint__}")
        #print(f'==> Scope.__del__ {type(self).__name__} {self.__joint__.tag} id={id(self):X} __joint__.id={id(self.__joint__):X}')
        pass
      
class Element(Scope):
    def __init__(self, joint):
        super().__init__(joint)        

class Section(Scope):
    def __init__(self, joint):
        super().__init__(joint)

class Cluster(Scope):
    def __init__(self, joint):
        super().__init__(joint)

class Decorator:
    def __init__(self, context, tag=None):
        self.context = context
        self.tag = tag
    def scope(self):
        return self.context.scope
    def __getattr__(self,name):
        return getattr(self.context.scope,name)
    def update(self, scope=None, **kwargs):
        if scope is None:
            scope = self.context.scope
        return scope.__joint__.__dict__.update(kwargs)
    def unbound(self, scope):
        return scope.__joint__.bond is None
    def realize(self, scope, *args, **kwargs):
        return scope.__joint__.realize(scope,*args,**kwargs)
    def free(self,scope):
        scope.__joint__.free(scope)


class Context:
    def __init__(self, tag = None):
        self.tag = tag
        self.scope_root = Scope(Joint(pool=self, base=None,mass=None,tool=None,tag='root'))
        self.scope = self.scope_root
        self.all = []
        self.line = []
        self.stack = []
        self.head = 0
        self.trail0 = {}
        self.trail = None
        self.stuff = None
        self.default_layer = None
    def take(self, builder):
        if builder.context is not self:
            return
        self.scope = builder.scope
        self.line.append(builder)
        self.head += 1
    def drop(self, builder):
        if builder.context is not self:
            return
        #### builder.context = None
        self.head -= 1
        self._build()
        self.scope = builder.scope.__joint__.base
    def select(self, scope):
        self.stack.append((self.scope,self.line,self.head,self.trail0,self.stuff))
        self.stuff = None
        self.trail = self.stack[-1][3]
        self.trail0 = {}
        self.scope = scope
        self.line = []
        self.head = 0
    def restore(self):
        self._build()
        self.scope, self.line, self.head, self.trail0, self.stuff = self.stack.pop() 
        self.trail = self.stack[-1][3] if self.stack else None
        return self.scope
    def _build(self):
        if self.head == 0 and self.line:
            line = self.line
            self.line = []
            for builder in line:
                save_scope = self.scope
                self.scope = builder.scope
                self.all.append(builder)
                builder.build()
                self.scope = save_scope
            if self.all:
                if builder is self.all[0]:
                    line = self.all
                    self.all = []
                    for builder in line:
                        save_scope = self.scope
                        with builder.scope:                        
                            builder.stage()

class ElementBuilder:
    def __init__(self, context=None, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs):
        self.context : Context = context
        # self.scope = Element(Joint(pool=context, base=scope, tag=intents, layer=layer))
        self.scope = None
        self.layer : layers.common.Layer = layer #self.scope.__joint__.layer
        self.base = scope 
        self.type = type
        self.intents = intents
        self.item = item
        self.helper = None
        self.name = None
        self.args = args
        self.kwargs = kwargs
        self.this = None
        # self.context.take(self)
    def __call__(self, item):
        if self.base is not None and hasattr(self.base,item.__name__):
            self.scope = getattr(self.base,item.__name__)
            self.building = False
        else:
            self.scope = Element(Joint(pool=self.context, base=self.base, tag=self.intents, layer=self.layer))
            self.building = True
        self.layer = self.scope.__joint__.layer
        self.context.take(self)
        target, wrapper, helper, name = self.scope.__joint__.element_wrap(self.scope,item.__name__,item)
        self.item = item
        self.helper = helper
        self.wrapper = wrapper
        self.name = name
        self.context.drop(self)
        return self.scope ### target            
    def __enter__(self) -> Scope:        
        self.layer.item_enter(self)
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.layer.item_leave(self)
    def stage(self):
        self.layer.item_stage(self)
    def build(self):
        scope = self.scope
        item = self.item
        helper = self.helper
        if self.building:
            this = self.layer.item_build(self)
            scope.__joint__.this = this
        else:
            this = scope.__joint__.this
        scope.__joint__.call = self.wrapper
        scope.__joint__.call2 = self.helper.wrapper
        self.this = this
        attrs = self.kwargs.get('attrs',None)
        if attrs:
            scope.__dict__.update(attrs)
        if item is not None:
            delegated_bindings = [k for k in self.kwargs if k in ('bind','bindings')]
            with self:
                if type(item) is str:                
                    pass ##print('build_end -> str -> ', item)    
                elif isinstance(item, types.FunctionType):
                    setattr(scope.__joint__.base,self.name,scope)
                    if not delegated_bindings:
                        helper(scope)
                else:
                    setattr(scope.__joint__.base,self.name,scope)
                    if not delegated_bindings:
                        if this is not None and helper is not None:
                            helper(this, scope)
            
class ElementDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        self.stack = []
    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__\n')
    def __call__(self, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs) -> ElementBuilder:
        if scope is None:
            scope = self.scope()
        self.update(scope, closest_layer=layer)
        return ElementBuilder(context=self.context, scope=scope, type=type, intents=intents, item=item, layer=layer, *args, **kwargs)
    def __enter__(self) -> Scope:        
        self.stack.append(self.context.scope)
        return self.context.scope
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.scope = self.stack.pop()
    def free(self, scope):
        return scope.__joint__.free(scope)

class SectionBuilder:
    def __init__(self, context=None, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs):
        self.context : Context = context
        self.scope = Section(Joint(pool=context, base=scope, tag=intents, layer=layer))
        self.layer : layers.common.layer = self.scope.__joint__.layer
        self.base = scope
        self.type = type
        self.intents = intents
        self.item = item
        self.helper = None
        self.name = None
        self.args = args
        self.kwargs = kwargs
        self.this = None
        self.context.take(self)        

    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__ {self.name}\n')

    def __call__(self, item):
        #print("ElementBuilder.__call___:",self.intents)
        # self.scope = Section(Joint(pool=self.context, base=self.base, tag=self.intents))
        #self.context.take(self)
        target, wrapper, helper, name = self.scope.__joint__.section_wrap(self.scope,item.__name__,item)
        self.item = item
        self.helper = helper
        self.wrapper = wrapper
        self.name = name
        self.context.drop(self)
        return self.scope 
    def __enter__(self) -> Scope:        
        return None
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    def build(self):
        scope = self.scope
        item = self.item
        this = self.layer.item_build(self)
        if  False:
            if type(self.type) is type:       
                this = self.type()     
            else:
                if type(item) is type:
                    this = item()     
                else:
                    this = item
        scope.__joint__.call = self.wrapper
        scope.__joint__.call2 = self.helper.wrapper
        scope.__joint__.this = this
        self.this = this
        attrs = self.kwargs.get('attrs',None)
        if attrs:
            scope.__dict__.update(attrs)
        setattr(scope.__joint__.base,self.name,scope)        
    def stage(self):
        pass

class SectionDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        self.stack = []
    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__\n')
    def __call__(self, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs) -> SectionBuilder:
        if scope is None:
            scope = self.scope()
        self.update(scope, closest_layer=layer)
        return SectionBuilder(context=scope.__joint__.pool, scope=scope, type=type, intents=intents, item=item, layer=layer, *args, **kwargs)
    def __enter__(self) -> Scope:        
        self.stack.append(self.context.scope)
        return self.context.scope
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.scope = self.stack.pop()
    def free(self, scope):
        return scope.__joint__.free(scope)
    
class ClusterBuilder:
    def __init__(self, context=None, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs):
        self.context : Context = context
        self.scope = Cluster(Joint(pool=context, base=scope, tag=intents, layer=layer))
        self.layer : layers.common.layer = self.scope.__joint__.layer
        self.base = scope
        self.type = type
        self.intents = intents
        self.item = item
        self.helper = None
        self.name = None
        self.args = args
        self.kwargs = kwargs
        self.this = None
        self.context.take(self)        

    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__ {self.name}\n')

    def __call__(self, item):
        #print("ElementBuilder.__call___:",self.intents)
        # self.scope = Cluster(Joint(pool=self.context, base=self.base, tag=self.intents))
        #self.context.take(self)
        target, wrapper, helper, name = self.scope.__joint__.cluster_wrap(self.scope,item.__name__,item)
        self.item = item
        self.helper = helper
        self.wrapper = wrapper
        self.name = name
        self.context.drop(self)
        return self.scope 
    def __enter__(self) -> Scope:        
        return None
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    def build(self):
        scope = self.scope
        item = self.item
        if type(self.type) is type:       
            this = self.type()     
        else:
            if type(item) is type:
                this = item()     
            else:
                this = item
        scope.__joint__.call2 = self.helper.wrapper
        scope.__joint__.call = self.wrapper
        scope.__joint__.this = this
        self.this = this
        attrs = self.kwargs.get('attrs',None)
        if attrs:
            scope.__dict__.update(attrs)
        setattr(scope.__joint__.base,self.name,scope)        
    def stage(self):
        pass

class ClusterDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        self.stack = []
    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__\n')
    def __call__(self, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs) -> ClusterBuilder:
        if scope is None:
            scope = self.scope()
        self.update(scope, closest_layer=layer)
        return ClusterBuilder(context=scope.__joint__.pool, scope=scope, type=type, intents=intents, item=item, layer=layer, *args, **kwargs)
    def __enter__(self) -> Scope:        
        self.stack.append(self.context.scope)
        return self.context.scope
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.scope = self.stack.pop()
    def free(self, scope):
        return scope.__joint__.free(scope)

class FeatureBuilder:
    def __init__(self, context=None, scope=None, host=None, type=None, intents=None, item=None, layer=None, overrides=None, *args, **kwargs):
        self.context : Context = context
        self.scope = Cluster(Joint(pool=context, base=scope, tag=intents, layer=layer, host=host, overrides=overrides))
        self.layer : layers.common.layer = self.scope.__joint__.layer
        self.base = scope
        self.type = type
        self.intents = intents
        self.item = item
        self.helper = None
        self.name = None
        self.args = args
        self.kwargs = kwargs
        self.this = None
        self.context.take(self)        

    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__ {self.name}\n')

    def __call__(self, item):
        #print("ElementBuilder.__call___:",self.intents)
        # self.scope = Cluster(Joint(pool=self.context, base=self.base, tag=self.intents))
        #self.context.take(self)
        target, wrapper, helper, name = self.scope.__joint__.feature_wrap(self.scope,item.__name__,item)
        self.item = item
        self.helper = helper
        self.wrapper = wrapper
        self.name = name
        self.context.drop(self)
        return self.scope 
    def __enter__(self) -> Scope:        
        return None
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    def build(self):
        scope = self.scope
        item = self.item
        if type(self.type) is type:       
            this = self.type()     
        else:
            if type(item) is type:
                this = item()     
            else:
                this = item
        scope.__joint__.call2 = self.helper.wrapper
        scope.__joint__.call = self.wrapper
        scope.__joint__.this = this
        self.this = this
        attrs = self.kwargs.get('attrs',None)
        if attrs:
            scope.__dict__.update(attrs)
        setattr(scope.__joint__.base,self.name,scope)        
    def stage(self):
        pass

class FeatureDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        self.stack = []
    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__\n')
    def __call__(self, scope=None, host=None, type=None, intents=None, item=None, layer=None, overrides=None, *args, **kwargs) -> FeatureBuilder:
        if scope is None:
            scope = self.scope()
        self.update(scope, closest_layer=layer)
        return FeatureBuilder(context=scope.__joint__.pool, scope=scope, host=host, type=type, intents=intents, item=item, layer=layer, 
                                overrides=overrides, *args, **kwargs)
    def __enter__(self) -> Scope:        
        self.stack.append(self.context.scope)
        return self.context.scope
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.scope = self.stack.pop()
    def free(self, scope):
        return scope.__joint__.free(scope)

    
class RoutineBuilder:
    def __init__(self, context, scope, host, *args, **kwargs):
        self.context : Context = context
        self.scope : Scope = scope
        self.layer : layers.common.Layer = self.scope.__joint__.layer
        self.host = host
        self.args = args
        self.kwargs = kwargs
        self.context.head += 1
        pass
    def __call__(self, item):
        self.context.head -= 1
        self.wrapper = self.scope.__joint__.routine_wrap(self.scope, self.host, item.__name__, item, **self.kwargs)       
        self.layer.item_bind(self)
        if self.kwargs.get('implicit',False):
            self.wrapper(self.scope)
        return self.wrapper

      
class RoutineDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        pass
    def __call__(self, scope=None, host=None, *args, **kwargs):
        if scope is None:
            scope = element.scope()
        return RoutineBuilder(context=scope.__joint__.pool,scope=scope,host=host,*args,**kwargs)
    def free(self, scope):
        pass

class BindingBuilder:
    def __init__(self, context=None, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs):
        self.context : Context = context
        self.scope :Scope = scope
        self.layer : layers.common.Layer = self.scope.__joint__.layer
        self.base = scope 
        self.type = type
        self.intents = intents
        self.item = item
        self.helper = None
        self.name = None
        self.args = args
        self.kwargs = kwargs
        self.this = None
        self.context.take(self)
    def __call__(self, item):
        target, wrapper, helper, name = self.scope.__joint__.binding_wrap(self.scope,item.__name__,item)
        self.item = item
        self.helper = helper
        self.wrapper = wrapper
        self.name = name
        self.context.drop(self)
        return self.scope ### target            
    def __enter__(self) -> Scope:        
        self.layer.item_enter(self)
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.layer.item_leave(self)
    def stage(self):
        self.layer.item_stage(self)
    def build(self):
        scope = self.scope
        item = self.item
        helper = self.helper
        this = self.layer.item_build(self)
        scope.__joint__.this = this
        #scope.__joint__.call = self.wrapper
        #scope.__joint__.bond = self.wrapper
        scope.__joint__.bond = scope.__joint__.call2

class BindingDecorator(Decorator):
    def __init__(self, context, tag=None):                    
        Decorator.__init__(self,context,tag)
        self.stack = []
    #def __del__(self):        
    #    print(f'\n===> {type(self).__name__} __del__\n')
    def __call__(self, scope=None, type=None, intents=None, item=None, layer=None, *args, **kwargs) -> ElementBuilder:
        if scope is None:
            scope = self.scope()
        self.update(scope, closest_layer=layer)
        return BindingBuilder(context=self.context, scope=scope, type=type, intents=intents, item=item, layer=layer, *args, **kwargs)
    def __enter__(self) -> Scope:        
        self.stack.append(self.context.scope)
        return self.context.scope
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.scope = self.stack.pop()
    def free(self, scope):
        return scope.__joint__.free(scope)



def decorator_set(tag=None):
    context = Context(tag)
    element = ElementDecorator(context,tag)
    section = SectionDecorator(context,tag)
    cluster = ClusterDecorator(context,tag)
    feature = FeatureDecorator(context,tag)
    routine = RoutineDecorator(context,tag)
    binding = BindingDecorator(context,tag)
    return element, section, cluster, feature, routine, binding

element, section, cluster, feature, routine, binding = decorator_set('default')


'''
    [Scope]
    [Section]
    [Element]
    [Spawned]
    [Routine]
    Context
    Layer
    ElementDecorator.__call__ -> ElementBuilder
        EmentBuilder.__call__
            self.scope.__joint__.element_wrap :> istanza i wrappers
            self.context.drop(self)
                foreach builder in line:
                    builder.build() :>
                        layer.element_build
                        helper(...)  :> - generato in element_wrap
                                        - è la funzione originale, in caso il decoratore sia collegato ad una funzione,
                                        - se invece l'elemento decorato non è una funzione ma un oggetto...
                    builder.stage() :> ideato come supporto per propagare i layout del Tk senza doverli per forza rispecificare
'''





from .guiXXx4_gui import *
from . import gui04x1_tkliterals as tkliterals
from . import gui06x2_tktraits as tktraits
from . import gui07x2_tktools as tktools
from . import gui08x2_tktypes as tktypes
from tkinter import font as tkfonts


from . import gui14x1_wintraits as wintraits
from . import gui15x1_wintypes as wintypes
from . import gui16x1_winapi as winapi
from . import gui17x2_wintools as wintools

from . import guiX3x3_types as types
from . import guiX2x3_tools as tools




from pyrite.monolith import element, section, routine
from rich.console import Console
import threading
import time
import pyrite
import queue

@element()
def console(con):
    con.console = Console()
    def main(con=con):
        con.console.print('HELLO')
        while True:
            m = con.qin.get()
            match m:
                case 'print', args, kwargs:
                    con.console.print(*args,**kwargs) 
        pass
    con.thread = threading.Thread(target=main,args=(con,),daemon=True)
    con.qin = queue.Queue()
    @routine(con)
    def start(con):
        con.thread.start()
    @routine(con)
    def print(con, *args, **kwargs):
        con.qin.put(('print',args,kwargs))
        pass


console.start()
console.print('HELLO2')
time.sleep(3)
console.print('HELLO3')
time.sleep(3)
pass



import re

class SmartTupleBuilder:
    def __init__(self):
        pass
    def __call__(self, **kwds):
        return kwds

smarttuple = SmartTupleBuilder()

class SmartClass:
    def __init__(self,**kwargs):
       self.__SmartClass__ = list(kwargs.keys())
       self.__dict__.update(kwargs) 
    def as_tuple(self):
        return tuple([self.__dict__[k] for k in self.__SmartClass__])
    def as_dict(self):
        return {k:self.__dict__[k] for k in self.__SmartClass__}

class SmartClassBuilder:
    def __call__(self, **kwargs):
        return SmartClass(**kwargs)
    def from_tree(self,root):
        def walker(node):
            if type(node) is dict:                
                v = SmartClass(**{kk:walker(xx) for kk,xx in node.items()})
            elif type(node) is list:
                v = [walker(xx) for xx in node]
            elif type(node) is tuple:
                v = (walker(xx) for xx in node)
            else:
                v = node
            return v
        return walker(root)
        
smartclass = SmartClassBuilder()



class SmartMatch:
    def __init__(self):
        self.lastmatch : re.Match = None
    def __call__(self, *args, **kwds):
        self.lastmatch = re.match(*args,**kwds)
        return self.lastmatch
    def __getattr__(self,name):
        return self.__getitem__(name)
    def __getitem__(self, key):
        if type(key) is tuple:
            _key, _cast, _default = key
            _val = self.lastmatch.groupdict().get(_key,None)
            if _val is not None:
                return _cast(_val)
            else:
                return _default
        else:            
            return self.lastmatch.groupdict().get(key,None)
    def __contains__(self, key):
        v = self.lastmatch.groupdict().get(key,None)
        return v is not None
    def keys(self):
        return self.lastmatch.groupdict().keys()
    def items(self):
        return self.lastmatch.groupdict().items()
    def values(self):
        return self.lastmatch.groupdict().values()
    def escape(self, **keys):
        return {k:v.encode('unicode-escape').decode() for k,v in self.items() if v is not None and (not keys or k in keys)}
        


class SmartMatchBuilder:
    def __call__(self, **kwargs):
        return SmartMatch(**kwargs) 
smartmatch = SmartMatchBuilder()
